/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   08.03.2015
* @brief  Class to manage access points and protocol handlers.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Class loads access points in runtime and links access points with handlers.
**/

#ifndef ACCESSPOINTMANAGER_H
#define ACCESSPOINTMANAGER_H

#include <QObject>
#include <QString>
#include <QMap>
#include <QPluginLoader>

#include "ssaphandler.h"
#include "sibconfig.h"

namespace cutesib {

/**
 * @brief The AccessPointManager class manages access points and handlers.
 */
class AccessPointManager : public QObject
{
    Q_OBJECT
public:
    //TODO: make method to load access points.
    explicit AccessPointManager(SibConfig *config, QObject *parent = 0);
    ~AccessPointManager();

    void load(QString accessPointsDir);
    void load(SibConfig *config);
    void unload(QString accessPointsId);
    void unloadAll();

    void setAccessPoints(SsapHandler *handler);

signals:

public slots:

private:
     SsapHandler *_ssapHandler_;
     QMap<QString, QPluginLoader*> _loaders;
     SibConfig *_config;
};

} // Namespace cutesib

#endif // ACCESSPOINTMANAGER_H
