/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   08.03.2015
* @brief  Main file to start CuteSIB.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Make some configuration actions and starts CuteSIB.
**/


#include <iostream>
#include <signal.h>


#include <QCoreApplication>
#include <QDir>

#include "sibinitializer.h"
#include "logger.h"


cutesib::SibInitializer initializer(SIB_CONFIG_PATH);



void handler(int sig)
{

    qDebug("\nquit the application (user request signal = %d).\n", sig);

    cutesib::Logger::message() << "\nHave a nice day and say Goodbye to " << SIB_NAME << "!" << endl;
    initializer.stop();
    QCoreApplication::quit();
}


void catchUnixSignals(const std::vector<int>& quitSignals,
                      const std::vector<int>& ignoreSignals = std::vector<int>()) {


    // all these signals will be ignored.
    for ( int sig : ignoreSignals )
        signal(sig, SIG_IGN);

    // each of these signals calls the handler (quits the QCoreApplication).
    for ( int sig : quitSignals )
        signal(sig, handler);
}







int main(int argc, char *argv[])
{
    using cutesib::Logger;

    QCoreApplication a(argc, argv);

    //TODO: make signals catching for windows
    catchUnixSignals({SIGQUIT, SIGINT, SIGTERM, SIGHUP});

    // Sets handler for debug messages.
    qInstallMessageHandler(Logger::messageHandler);

    Logger::message() << "Starting "<< SIB_NAME << " version " <<  SIB_VERSION << endl;


    if (initializer.initialize() < 0) {
        Logger::message() << "Initialization failed." << endl;
        return 0;
    }

    initializer.start();

    return a.exec();
}
