#include "sib_data.h"
#include "sib_operations.h"
#include "sib_reasoning.h"
#include "sib_utilities.h"
/*
int file_exists(gchar *filename)
{
    FILE * file;
    if (file = fopen(filename, "r"))
    {
        fclose(file);
        return 1;
    }
    return 0;
}*/

void add_triplelist_in_submodel( sib_data_structure* p, ssap_sib_message * rsp, GSList * triplelist)

{

    librdf_statement * statement;
    ssTriple_t * t;

    while (triplelist !=NULL)
    {
        t= (ssTriple_t *) triplelist->data;
        statement=librdf_new_statement(rsp->subworld);

        librdf_statement_set_subject(statement,
                librdf_new_node_from_uri_string(rsp->subworld,  t->subject));

        librdf_statement_set_predicate(statement,
                librdf_new_node_from_uri_string(rsp->subworld, t->predicate));

        if (! t->objType) {
            //URI:  t->objType==0
            librdf_statement_set_object(statement,
                    librdf_new_node_from_uri_string(rsp->subworld,  t->object));
        }
        else
        {
            //Literal:      t->objType==1
            librdf_statement_set_object(statement,
                    librdf_new_node_from_literal(rsp->subworld, t->object, NULL, 0));

        }

        //qDebug("add_triplelist_in_submodel ADDING\n");
        //librdf_statement_print(statement, stdout);
        //qDebug("\n");

        librdf_model_add_statement(rsp->submodel, statement);
        librdf_free_statement(statement);

        triplelist=triplelist->next;
    }
}

void remove_triplelist_in_submodel( sib_data_structure* p, ssap_sib_message * rsp, GSList * triplelist)
{

    librdf_statement * statement;
    ssTriple_t * t;

    while (triplelist !=NULL)
    {

        t= (ssTriple_t *) triplelist->data;
        statement=librdf_new_statement(rsp->subworld);

        librdf_statement_set_subject(statement,
                librdf_new_node_from_uri_string(rsp->subworld,  t->subject));

        librdf_statement_set_predicate(statement,
                librdf_new_node_from_uri_string(rsp->subworld, t->predicate));

        if (! t->objType) {
            //URI:  t->objType==0
            librdf_statement_set_object(statement,
                    librdf_new_node_from_uri_string(rsp->subworld,  t->object));
        }
        else
        {
            //Literal:      t->objType==1
            librdf_statement_set_object(statement,
                    librdf_new_node_from_literal(rsp->subworld, t->object, NULL, 0));

        }

        //qDebug("remove_triplelist_in_submodel REMOVING\n");
        //librdf_statement_print(statement, stdout);
        //qDebug("\n");

        librdf_model_remove_statement(rsp->submodel, statement);
        librdf_free_statement(statement);

        triplelist=triplelist->next;
    }
}

void ssFreeTriple( ssTriple_t *triple)
{
  if(!triple)
    return; //nothing to do

  if (triple->subject && triple->subject!=ssMATCH_ANY) g_free (triple->subject);
  if (triple->predicate && triple->predicate!=ssMATCH_ANY) g_free (triple->predicate);
  if (triple->object && triple->object!=ssMATCH_ANY) g_free (triple->object);

  g_free (triple);
}

void ssFreeTripleList (GSList **tripleList)
{
  if (!tripleList || !*tripleList)
    return; //nothing to do

  GSList *list = *tripleList;
  ssTriple_t *triple;
  while (list)
    {
      triple = (ssTriple_t *) list->data;
      list = g_slist_remove(list,triple);
      ssFreeTriple(triple);
    }
  *tripleList = NULL;
}

int client_receive_message(char* message, NodeMsgContent_t* msg)
{
    gint rtmp;
    gint bytes_handled = 0;
    gint bytecountnew = 0;
    ssStatus_t status;
    gint index = 0;
    gboolean finished = FALSE;

    int len = strlen(message);
    int remaining_len = len;

    while(!finished)
    {
        if(remaining_len)
        {
            index = len - remaining_len;

            bytes_handled += remaining_len;
            qDebug("Parsing section; index:%d, remaining_len: %d, bytes_handled: %d\n",
                    index, remaining_len, bytes_handled);

            status = parseSSAPmsg_section (msg,
                    (gchar*) (message + index),
                    remaining_len,
                    0);

            if(status == ss_StatusOK)
            {
                int not_handled;
                bytecountnew = parseSSAPmsg_parsedbytecount(msg);
                not_handled = bytes_handled - bytecountnew;
                qDebug("Parse finished successfully, bytecount (%d), not_handled (%d)\n", bytecountnew, not_handled);
                remaining_len = not_handled;
                bytes_handled = bytecountnew;
                finished = TRUE;
            }
            else if(status == ss_ParsingInProgress)
            {
                len = 0;
                remaining_len = 0;
            }
            else
            {
                qWarning("Parse error %d\n", status);
                finished = TRUE;
                bytes_handled = -1;
            }
        }


    }
    return bytes_handled;
}

void listener_handle_request(NodeMsgContent_t *msg, sib_data_structure* sib_data)
{
    switch( parseSSAPmsg_get_name(msg))
    {

    case MSG_N_NSET:
        qWarning("message name not set...\n");
        break;
    /* case MSG_N_JOIN:
        listener_handle_join(msg);
        break;
    case MSG_N_LEAVE:
        listener_handle_leave(msg);
        break;*/
    case MSG_N_INSERT:
        listener_handle_insert(msg, sib_data);
        break;
    case MSG_N_UPDATE:
        listener_handle_update(msg, sib_data);
        break;
    case MSG_N_REMOVE:
        listener_handle_remove(msg, sib_data);
        break;
    case MSG_N_QUERY:
        char* res;
        res = listener_handle_query(msg, sib_data);
        break;
    case MSG_N_SUBSCRIBE:
        listener_handle_subscribe(msg, sib_data);
        break;
    case MSG_N_UNSUBSCRIBE:
        listener_handle_unsubscribe(msg, sib_data);
        break;

    default:

        qWarning("Unknown request: %d\n", parseSSAPmsg_get_name(msg));
        break;
    }
}

char* listener_handle_insert(NodeMsgContent_t *msg, sib_data_structure* sib_data)
{
    ssStatus_t status;
    ssBufDesc_t *bd = NULL;
    qDebug("Got Insert request\n");

    guchar* nodeid;
    guchar* spaceid;
    gint msgnumber;

    ssap_sib_message* rsp_msg = NULL;

    EncodingType encoding;
    if (parseSSAPmsg_get_graphStyle(msg) == MSG_G_TMPL )
    {	encoding = EncodingM3XML;
    }
    else if (parseSSAPmsg_get_graphStyle(msg) == MSG_G_RDF )
    {	encoding = EncodingRDFXML;
    }
    else
    {	encoding =EncodingInvalid;
    }

    nodeid = (guchar *)parseSSAPmsg_get_nodeid(msg);
    spaceid = (guchar *)parseSSAPmsg_get_spaceid(msg);
    msgnumber = parseSSAPmsg_get_msgnumber(msg);

    guchar* bnodes = NULL;
    if (encoding == EncodingInvalid)
    {
        qWarning("Could not send insert request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));

        status = ss_SIBFailNotImpl;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateInsertResponse(bd,
                                                          (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                          (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                          parseSSAPmsg_get_msgnumber(msg),
                                                          (msgStatus_t) status,
                                                          NULL))
        {
            ssBufDesc_free(&bd);
        }
    }
    else if ((rsp_msg = m3_insert(sib_data,
                                 (gchar*) parseSSAPmsg_get_spaceid(msg),
                                 (gchar*) parseSSAPmsg_get_nodeid(msg),
                                 parseSSAPmsg_get_msgnumber(msg),
                                 encoding,
                                 parseSSAPmsg_get_insert_graph(msg))) == NULL)
    {
        status = ss_OperationFailed;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateInsertResponse(bd,
                                                          (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                          (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                          parseSSAPmsg_get_msgnumber(msg),
                                                          (msgStatus_t) status,
                                                          NULL))
        {
            ssBufDesc_free(&bd);
        }
    }
    else
    {


        NodeMsgContent_t* rsp = nodemsgcontent_new_insert_rsp(spaceid, nodeid, msgnumber, rsp_msg->status, (guchar*) rsp_msg->bnodes_str);

        g_free(rsp_msg->bnodes_str);
        g_free(rsp_msg);

        if(rsp)
        {
            bd = ssBufDesc_new();
            if( ss_StatusOK != ssBufDesc_CreateInsertResponse(bd,
                                                              (guchar *) parseSSAPmsg_get_nodeid(rsp),
                                                              (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                              parseSSAPmsg_get_msgnumber(rsp),
                                                              parseSSAPmsg_get_msg_status(rsp),
                                                              (guchar *)parseSSAPmsg_get_insert_graph(rsp)) )
            {
                ssBufDesc_free(&bd);
            }

            nodemsgcontent_free(&rsp);
        }
    }

    //}
    if(bd)
    {
        char* message = strdup((char*) ssBufDesc_GetMessage(bd));
        ssBufDesc_free(&bd);

        return message;
    }

    return NULL;
}

char* listener_handle_remove(NodeMsgContent_t *msg, sib_data_structure* sib_data)
{
    ssStatus_t status;
    ssBufDesc_t *bd = NULL;
    qDebug("Got Remove request\n");

    gchar* nodeid;
    gchar* spaceid;
    gint msgnumber;


    //EncodingType encoding = ( parseSSAPmsg_get_graphStyle(msg) == MSG_G_TMPL ? EncodingM3XML : EncodingInvalid);

    EncodingType encoding;
    if  (parseSSAPmsg_get_graphStyle(msg) == MSG_G_TMPL )
    {
        encoding = EncodingM3XML;
    }
    else if (parseSSAPmsg_get_graphStyle(msg) == MSG_G_RDF )
    {
        encoding = EncodingRDFXML;
    }
    else
    {
        encoding =EncodingInvalid;
    }

    nodeid = (gchar*) parseSSAPmsg_get_nodeid(msg);
    spaceid = (gchar*) parseSSAPmsg_get_spaceid(msg);
    msgnumber = parseSSAPmsg_get_msgnumber(msg);


    if((encoding == EncodingInvalid) ||
            (m3_remove(sib_data, spaceid, nodeid, msgnumber, encoding, parseSSAPmsg_get_remove_graph(msg)) != ss_StatusOK))
    {
        qWarning("Could not send remove request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));

        if( encoding == EncodingInvalid)
            status = ss_SIBFailNotImpl;
        else
            status = ss_OperationFailed;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateRemoveResponse(bd,
                                                          (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                          (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                          parseSSAPmsg_get_msgnumber(msg),
                                                          (msgStatus_t) status))
        {
            ssBufDesc_free(&bd);
        }
    }
    else
    {
        NodeMsgContent_t* rsp = nodemsgcontent_new_remove_rsp((guchar*) spaceid, (guchar*) nodeid, msgnumber, ss_StatusOK);
        if(rsp)
        {
            bd = ssBufDesc_new();
            if( ss_StatusOK != ssBufDesc_CreateRemoveResponse(bd,
                                                              (guchar *)parseSSAPmsg_get_nodeid(rsp),
                                                              (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                              parseSSAPmsg_get_msgnumber(rsp),
                                                              parseSSAPmsg_get_msg_status(rsp) ) )
            {
                ssBufDesc_free(&bd);
            }

            nodemsgcontent_free(&rsp);
        }
    }

    //}

    if(bd)
    {
        char* message = strdup((char*) ssBufDesc_GetMessage(bd));
        ssBufDesc_free(&bd);

        return message;
    }

    return NULL;
}

char* listener_handle_update(NodeMsgContent_t *msg, sib_data_structure *sib_data)
{
    ssStatus_t status;
    ssBufDesc_t *bd = NULL;
    qDebug("Got update request\n");

    gchar* nodeid;
    gchar* spaceid;
    gint msgnumber;

    ssap_sib_message* rsp_msg = NULL;

 //EncodingType encoding = ( parseSSAPmsg_get_graphStyle(msg) == MSG_G_TMPL ? EncodingM3XML : EncodingInvalid);
    EncodingType encoding;

    if  (parseSSAPmsg_get_graphStyle(msg) == MSG_G_TMPL )
    {
        encoding = EncodingM3XML;
    }
    else if (parseSSAPmsg_get_graphStyle(msg) == MSG_G_RDF )
    {
        encoding = EncodingRDFXML;
    }
    else
    {
        encoding =EncodingInvalid;
    }

    nodeid = (gchar*) parseSSAPmsg_get_nodeid(msg);
    spaceid = (gchar*) parseSSAPmsg_get_spaceid(msg);
    msgnumber = parseSSAPmsg_get_msgnumber(msg);
    char* insert_str = (char *) parseSSAPmsg_get_insert_graph(msg);
    char* remove_str = (char *) parseSSAPmsg_get_remove_graph(msg);


    if((encoding == EncodingInvalid) ||
       ((rsp_msg = m3_update(sib_data, spaceid, nodeid, msgnumber, encoding,
                            insert_str, remove_str)) == NULL))
    {
        qDebug("Could not send update request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));

        if(encoding == EncodingInvalid)
            status = ss_SIBFailNotImpl;
        else
            status = ss_OperationFailed;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateUpdateResponse(bd,
                                                          (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                          (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                          parseSSAPmsg_get_msgnumber(msg),
                                                          (msgStatus_t) status,
                                                          NULL))
        {
            ssBufDesc_free(&bd);
        }
    }
    else
    {

        NodeMsgContent_t* rsp = nodemsgcontent_new_update_rsp((guchar*) spaceid, (guchar*) nodeid, msgnumber, rsp_msg->status, (guchar*) rsp_msg->bnodes_str);

        g_free(rsp_msg->bnodes_str);
        g_free(rsp_msg);

        if(rsp)
        {
            bd = ssBufDesc_new();
            if( ss_StatusOK != ssBufDesc_CreateUpdateResponse(bd,
                                                              (guchar *)parseSSAPmsg_get_nodeid(rsp),
                                                              (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                              parseSSAPmsg_get_msgnumber(rsp),
                                                              parseSSAPmsg_get_msg_status(rsp),
                                                              (guchar *)parseSSAPmsg_get_M3XML(rsp)) )
            {
                ssBufDesc_free(&bd);
            }

            nodemsgcontent_free(&rsp);
        }
    }
    //}

    if(bd)
    {
        char* message = strdup((char*) ssBufDesc_GetMessage(bd));
        ssBufDesc_free(&bd);

        return message;
    }

    return NULL;
}

char* listener_handle_query(NodeMsgContent_t *msg, sib_data_structure *sib_data)
{
    ssStatus_t status = ss_StatusOK;
    ssBufDesc_t *bd = NULL;
    qDebug("Got query request\n");

    gchar* nodeid;
    gchar* spaceid;
    gint msgnumber;


    QueryType qtype;
    switch(parseSSAPmsg_get_queryStyle( msg) )
    {
    case   MSG_Q_NSET:
    case   MSG_Q_N3:
        qtype = QueryTypeInvalid;
        break;
    case   MSG_Q_TMPL:
        qtype=QueryTypeTemplate;
        break;
    case   MSG_Q_SPRQL:
        qtype=QueryTypeSPARQLSelect;
        break;
    case   MSG_Q_WQL_VALUES:
        qtype=QueryTypeWQLValues;
        break;
    case   MSG_Q_WQL_NODETYPES:
        qtype=QueryTypeWQLNodeTypes;
        break;
    case   MSG_Q_WQL_RELATED:
        qtype=QueryTypeWQLRelated;
        break;
    case   MSG_Q_WQL_ISTYPE:
        qtype=QueryTypeWQLIsType;
        break;
    case   MSG_Q_WQL_ISSUBTYPE:
        qtype=QueryTypeWQLIsSubType;
        break;
    default:
        qtype = QueryTypeInvalid;
        break;
    }

    nodeid = (gchar *)parseSSAPmsg_get_nodeid(msg);
    spaceid = (gchar *)parseSSAPmsg_get_spaceid(msg);
    msgnumber = parseSSAPmsg_get_msgnumber(msg);

    /* REMOVE
            if( (qtype == QueryTypeInvalid) ||
                    ( sib_object_send_query( listener->object,
                            (guchar *)parseSSAPmsg_get_spaceid(msg),
                            (guchar *)parseSSAPmsg_get_nodeid(msg),
                            parseSSAPmsg_get_msgnumber(msg),
                            qtype,
                            (guchar *)parseSSAPmsg_get_M3XML(msg) )  <0) )
                            */
    ssap_sib_message* rsp_msg = NULL;

    if ((qtype == QueryTypeInvalid) ||
            (rsp_msg = m3_query_spql_upd(sib_data, spaceid, nodeid, msgnumber, qtype, (char *)parseSSAPmsg_get_M3XML(msg))) == NULL)
    {
        qDebug("Could send query request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));

        if( qtype == QueryTypeInvalid)
            status = ss_SIBFailNotImpl;
        else
            status = ss_OperationFailed;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateQueryResponse(bd,
                                                         (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                         (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                         parseSSAPmsg_get_msgnumber(msg),
                                                         (msgStatus_t) status,
                                                         NULL))
        {
            ssBufDesc_free(&bd);
        }
    }
    else
    {

        NodeMsgContent_t* rsp = nodemsgcontent_new_query_rsp((guchar*) spaceid, (guchar*) nodeid, msgnumber, rsp_msg->status, (guchar*) rsp_msg->results_str);

        g_free(rsp_msg->results_str);
        g_free(rsp_msg);

        if (rsp)
        {

            bd = ssBufDesc_new();
            if( ss_StatusOK != ssBufDesc_CreateQueryResponse(bd,
                                                             (guchar *)parseSSAPmsg_get_nodeid(rsp),
                                                             (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                             parseSSAPmsg_get_msgnumber(rsp),
                                                             parseSSAPmsg_get_msg_status(rsp),
                                                             (guchar *)parseSSAPmsg_get_M3XML(rsp)) )
            {
                ssBufDesc_free(&bd);
            }

            nodemsgcontent_free(&rsp);
        }
    }
    //}
    if(bd)
    {
        char* message = strdup((char*) ssBufDesc_GetMessage(bd));
        ssBufDesc_free(&bd);

        return message;
    }

    return NULL;
}

char* listener_handle_subscribe(NodeMsgContent_t *msg, sib_data_structure *sib_data)
{

    ssStatus_t status = ss_StatusOK;

    //Waiter *subwaiter = NULL;
    ssBufDesc_t *bd = NULL;
    gboolean subscription_finished = FALSE;
    qDebug("Got subscribe request\n");

    int optval=0;
    int optlen;
    int errorsock;

    int n =0;

    char* nodeid=	NULL;
    char* spaceid= NULL;
    char* subid=	NULL;


    NodeMsgContent_t* rsp = NULL;

    QueryType qtype;
    switch(parseSSAPmsg_get_queryStyle( msg) )
    {
    case   MSG_Q_NSET:
    case   MSG_Q_N3:
        qtype = QueryTypeInvalid;
        break;
    case   MSG_Q_TMPL:
        qtype=QueryTypeTemplate;
        break;
    case   MSG_Q_SPRQL:
        qtype=QueryTypeSPARQLSelect;
        break;
    case   MSG_Q_WQL_VALUES:
        qtype=QueryTypeWQLValues;
        break;
    case   MSG_Q_WQL_NODETYPES:
        qtype=QueryTypeWQLNodeTypes;
        break;
    case   MSG_Q_WQL_RELATED:
        qtype=QueryTypeWQLRelated;
        break;
    case   MSG_Q_WQL_ISTYPE:
        qtype=QueryTypeWQLIsType;
        break;
    case   MSG_Q_WQL_ISSUBTYPE:
        qtype=QueryTypeWQLIsSubType;
        break;
    default:
        qtype = QueryTypeInvalid;
        break;
    }

    /*
            if( (qtype == QueryTypeInvalid) ||
                    ( sib_object_send_subscribe( listener->object,
                            (guchar *)parseSSAPmsg_get_spaceid(msg),
                            (guchar *)parseSSAPmsg_get_nodeid(msg),
                            parseSSAPmsg_get_msgnumber(msg),
                            qtype,
                            (guchar *)parseSSAPmsg_get_M3XML(msg) )  <0) )*/

    GCond *sub_cond;
    GMutex *sub_mutex;

    sub_mutex = g_mutex_new();
    sub_cond = g_cond_new();

    subscribe_param* s;
    s = g_new0(subscribe_param, 1);

    //s->listener = listener;
    s->spaceid = (guchar *)parseSSAPmsg_get_spaceid(msg);
    s->nodeid = (guchar *)parseSSAPmsg_get_nodeid(msg);
    s->msgnumber = parseSSAPmsg_get_msgnumber(msg);
    s->qtype = qtype;
    s->query_str = (guchar *)parseSSAPmsg_get_M3XML(msg);
    s->subscription_id = NULL;
    s->results = NULL;
    s->sub_cond = sub_cond;
    s->sib_data = sib_data;

    GError* gerror = NULL;

    //g_thread_new("subscribe", subscribe_thread, (gpointer) s);
    GThread *t = g_thread_try_new("subscribe", subscribe_thread, (gpointer) s, &gerror);
    if (t == NULL) {
        qDebug("ERROR: %s %d", gerror->message, gerror->code);
    }

    //m3_subscribe(sib_data, (gchar*) s->spaceid, (gchar*) s->nodeid, s->msgnumber,
    //           s->qtype, (const char*) s->query_str, &(s->subscription_id), &(s->results), s->sub_cond);

    qDebug(" ********** BEFORE COND WAIT\n");
    g_mutex_lock(sub_mutex);
    g_cond_wait(sub_cond, sub_mutex);
    qDebug(" ********** AFTER COND WAIT\n");

    /*
            if( (qtype == QueryTypeInvalid) ||
                    (m3_subscribe(listener, (guchar *)parseSSAPmsg_get_spaceid(msg),
                     (guchar *)parseSSAPmsg_get_nodeid(msg),
                     parseSSAPmsg_get_msgnumber(msg),
                                  qtype,
                                  (guchar *)parseSSAPmsg_get_M3XML(msg),
                                  subscription_id, results) < 0))
                                  */
    if( (qtype == QueryTypeInvalid))
    {
        qWarning("Could not send subscribe request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));
        //listener_remove_request(listener, (guchar *)parseSSAPmsg_get_nodeid( msg), parseSSAPmsg_get_msgnumber( msg));
        if( qtype == QueryTypeInvalid)
            status = ss_SIBFailNotImpl;
        else
            status = ss_OperationFailed;


        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateSubscribeResponse(bd,
                                                             (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                             (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                             parseSSAPmsg_get_msgnumber(msg),
                                                             (msgStatus_t) status,
                                                             NULL,
                                                             NULL))
        {
            ssBufDesc_free(&bd);
        }
        subscription_finished = TRUE;
    }
    else
    {
        //waiter_wait_for_finished(waiter);
        //if(waiter->rsp)
        qDebug("DONE\n");
        qDebug("%s\n", s->subscription_id);

        rsp = nodemsgcontent_new_subscribe_rsp((guchar *)parseSSAPmsg_get_spaceid(msg),
                                               (guchar *)parseSSAPmsg_get_nodeid(msg),
                                               parseSSAPmsg_get_msgnumber(msg),
                                               status, (guchar*) s->subscription_id,
                                               (guchar*) s->results);
        if (rsp)
        {
            //subwaiter = waiter_new(client->fd);
            bd = ssBufDesc_new();

            /* WARNING: TODO check on wrong response
                     *
                    if( listener_add_subscription(listener,  (guchar *)parseSSAPmsg_get_subscriptionid(rsp),subwaiter) < 0)
                    {

                        qDebug("Could not add new subscription w/ id : %s\n", parseSSAPmsg_get_subscriptionid(rsp));
                        //waiter_destroy(subwaiter);
                        if( ss_StatusOK != ssBufDesc_CreateSubscribeResponse(bd,
                                (guchar *)parseSSAPmsg_get_nodeid(msg),
                                (guchar *)parseSSAPmsg_get_spaceid(msg),
                                parseSSAPmsg_get_msgnumber(msg),
                                status,
                                NULL,
                                NULL))
                        {
                            ssBufDesc_free(&bd);
                        }
                    }
                    else*/
            if (ss_StatusOK != ssBufDesc_CreateSubscribeResponse(
                        bd,
                        (guchar *)parseSSAPmsg_get_nodeid(rsp),
                        (guchar *)parseSSAPmsg_get_spaceid(rsp),
                        parseSSAPmsg_get_msgnumber(rsp),
                        parseSSAPmsg_get_msg_status(rsp),
                        (guchar *)parseSSAPmsg_get_subscriptionid(rsp),
                        (guchar *)parseSSAPmsg_get_M3XML(rsp)))
            {
                ssBufDesc_free(&bd);
            }

            //////////FIXED PARAMETERS//////////
            nodeid = (char*) malloc(strlen(parseSSAPmsg_get_nodeid(rsp))+1);
            strncpy(nodeid, parseSSAPmsg_get_nodeid(rsp),strlen(parseSSAPmsg_get_nodeid(rsp)) );
            nodeid[strlen(parseSSAPmsg_get_nodeid(rsp))] = '\0';

            spaceid = (char*) malloc(strlen(parseSSAPmsg_get_spaceid(rsp))+1);
            strcpy(spaceid, parseSSAPmsg_get_spaceid(rsp) );
            spaceid[strlen(parseSSAPmsg_get_spaceid(rsp))] = '\0';

            subid = (char*) malloc(strlen(parseSSAPmsg_get_subscriptionid(rsp))+1);
            strncpy(subid, parseSSAPmsg_get_subscriptionid(rsp),strlen(parseSSAPmsg_get_subscriptionid(rsp)) );
            subid[strlen(parseSSAPmsg_get_subscriptionid(rsp))] = '\0';
            ////////////////////////////////////

        }
    }

    if(bd && rsp && (parseSSAPmsg_get_msg_status(rsp)==2))
    {
        /* WARNING: TODO check on send failed
         *
        errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
        if (optval!=0)
        {
            qDebug("Send failed\n");
            subscription_finished = TRUE;
        }
        else if( client_send_message(client->fd,
                (guchar *)ssBufDesc_GetMessage(bd),
                ssBufDesc_GetMessageLen(bd)) < 0 )
        {
            qDebug("Send failed\n");
            subscription_finished = TRUE;
        }


        */
        //listener_remove_request(listener, (guchar *)parseSSAPmsg_get_nodeid( msg), parseSSAPmsg_get_msgnumber(msg));

        nodemsgcontent_free(&rsp);

        if(bd)
        {
            char* message = strdup((char*) ssBufDesc_GetMessage(bd));
            ssBufDesc_free(&bd);

            free (nodeid);
            free (subid);
            free (spaceid);

            return message;
        }

        //ssBufDesc_free(&bd);

    }

    free (nodeid);
    free (subid);
    free (spaceid);
        // WARNING: TODO SEND SUBSCRIPTION INDICATION RESPONSE

        /*
        while ( !subscription_finished  )
        {
            qDebug("Waiting subscription_ind\n");
            //waiter_wait_for_finished( subwaiter );
            qDebug("got something \n");


            bd = ssBufDesc_new();
            if(subwaiter->rsp)
            {
                if( ( parseSSAPmsg_get_name(subwaiter->rsp) == MSG_N_SUBSCRIBE)  &&
                        ( parseSSAPmsg_get_type(subwaiter->rsp) == MSG_T_IND)  &&
                        ( ss_StatusOK == ssBufDesc_CreateSubscriptionIndMessage(bd,
                                (guchar *)parseSSAPmsg_get_nodeid(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_spaceid(subwaiter->rsp),
                                parseSSAPmsg_get_msgnumber(subwaiter->rsp),
                                parseSSAPmsg_get_update_sequence (subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_results_added(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_results_removed(subwaiter->rsp) ) ))
                {
                    qDebug("Got Subscription ind, id: %s\n",parseSSAPmsg_get_subscriptionid(subwaiter->rsp));

                    errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
                    if (optval!=0)
                    {
                        qDebug("Send failed by optval\n");

                        ///////////////////AUTO UNSUB////////////////////
                        if( sib_object_send_unsubscribe(listener->object,
                                (guchar *)parseSSAPmsg_get_spaceid(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_nodeid(subwaiter->rsp),
                                parseSSAPmsg_get_msgnumber(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp)) <0)
                        {
                            qDebug("Could not send unsubscribe request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber(msg ));
                        }
                        ///////////////////AUTO UNSUB////////////////////

                        listener_remove_subscription(listener, (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp));
                        subscription_finished = TRUE;
                    }
                    else if( client_send_message(subwaiter->fd,
                            (guchar *)ssBufDesc_GetMessage(bd),
                            ssBufDesc_GetMessageLen(bd)) < 0 )
                    {
                        qDebug("Send failed\n");

                        ///////////////////AUTO UNSUB////////////////////
                        if( sib_object_send_unsubscribe(listener->object,
                                (guchar *)parseSSAPmsg_get_spaceid(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_nodeid(subwaiter->rsp),
                                parseSSAPmsg_get_msgnumber(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp)) <0)
                        {
                            qDebug("Could not send unsubscribe request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber(msg ));
                        }

                        ///////////////////AUTO UNSUB////////////////////
                        listener_remove_subscription(listener, (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp));
                        subscription_finished = TRUE;

                    }
                    else
                    {
                        nodemsgcontent_free( &(subwaiter->rsp) );
                    }

                }
                else if ( (parseSSAPmsg_get_name(subwaiter->rsp) == MSG_N_UNSUBSCRIBE)  &&
                        ( parseSSAPmsg_get_type(subwaiter->rsp) == MSG_T_CNF)  &&
                        ( ss_StatusOK == ssBufDesc_CreateUnsubscribeResponse(bd,
                                (guchar *)parseSSAPmsg_get_nodeid(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_spaceid(subwaiter->rsp),
                                parseSSAPmsg_get_msgnumber(subwaiter->rsp),
                                parseSSAPmsg_get_msg_status(subwaiter->rsp),
                                (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp) )))
                {
                    qDebug("Got Unsubscribe cnf, id %s\n",parseSSAPmsg_get_subscriptionid(subwaiter->rsp));

                    errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
                    if (optval!=0)
                    {
                        qDebug("Send failed by optval\n");
                    }
                    else if( client_send_message(subwaiter->fd,
                            (guchar *)ssBufDesc_GetMessage(bd),
                            ssBufDesc_GetMessageLen(bd)) < 0 )
                    {
                        qDebug("Send failed\n");
                    }

                    subscription_finished = TRUE;
                    listener_remove_subscription(listener, (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp));
                }
                else
                {
                    qDebug("Got invalid response , id %s\n",parseSSAPmsg_get_subscriptionid(subwaiter->rsp));
                    listener_remove_subscription(listener, (guchar *)parseSSAPmsg_get_subscriptionid(subwaiter->rsp));
                    subscription_finished = TRUE;
                }
            }
            else
            {
                qDebug("Exploring socket\n");

                errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
                qDebug("Socket broken=%d \n",optval);
                qDebug("Socket broken_ck2=%d \n",errorsock);

                if ((optval!=0)||(n==-1))
                {
                    qDebug("\n Socket broken.\n");

                    ///////////////////AUTO UNSUB///////////////////
                    if( sib_object_send_unsubscribe(listener->object,
                            (guchar *)spaceid,
                            (guchar *)nodeid,
                            1,
                            (guchar *)subid) <0)
                    {
                        qDebug("Could not send unsubscribe request w/ msgid : %d\n",1);
                    }
                    ///////////////////AUTO UNSUB////////////////////

                    listener_remove_subscription(listener, subid);
                    subscription_finished = TRUE;

                    close(client->fd);

                    ssBufDesc_free(&bd);

                    free (nodeid);
                    free (subid);
                    free (spaceid);

                    return 0;
                }
                //////////////////////////////////////////////////////

                //////One byte SOCKET EXPLORER/////////
                n=send(subwaiter->fd, " ", 1, MSG_NOSIGNAL);


                ///////////////////////////////////////
                qDebug("Exploredt\n");

                errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
                qDebug("Socket broken=%d \n",optval);


                //qDebug (" spaceid %s \n nodeid %s \n subid %s\n",spaceid,nodeid,subid);

                if ((optval!=0)||(n==-1))
                {
                    qDebug("\n Socket broken\n");

                    ///////////////////AUTO UNSUB///////////////////
                    if( sib_object_send_unsubscribe(listener->object,
                            (guchar *)spaceid,
                            (guchar *)nodeid,
                            1,
                            (guchar *)subid) <0)
                    {
                        qDebug("Could not send unsubscribe request w/ msgid : %d\n",1);
                    }
                    ///////////////////AUTO UNSUB////////////////////

                    listener_remove_subscription(listener, subid);
                    subscription_finished = TRUE;

                    ssBufDesc_free(&bd);
                    close(client->fd);

                    free (nodeid);
                    free (subid);
                    free (spaceid);

                    return 0;

                }
                //////////////////////////////////////////////////////

            }
            ssBufDesc_free(&bd);
        }
    }*/
    /*
    else  if(bd && rsp && (parseSSAPmsg_get_msg_status(rsp)!=2))
    {


        listener_remove_subscription(listener, subid);
        //qDebug("\nbad subscribe..\n");
        //qDebug ("value %d \n",parseSSAPmsg_get_msg_status(waiter->rsp) );

        errorsock=getsockopt(subwaiter->fd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
        if (optval!=0)
        {
            qDebug("Send failed\n");
        }
        else if( client_send_message(client->fd,
                (guchar *)ssBufDesc_GetMessage(bd),
                ssBufDesc_GetMessageLen(bd)) < 0 )
        {
            qDebug("Send failed\n");
        }

        listener_remove_request(listener, (guchar *)parseSSAPmsg_get_nodeid( msg), parseSSAPmsg_get_msgnumber(msg));
        ssBufDesc_free(&bd);

        close(client->fd);

    }
*/

    return NULL;
}

char* listener_handle_unsubscribe(NodeMsgContent_t *msg, sib_data_structure *sib_data)
{
    ssStatus_t status;

    NodeMsgContent_t* rsp = NULL;
    ssBufDesc_t *bd = NULL;
    qDebug("Got unsubscribe request\n");

    /*
            if( sib_object_send_unsubscribe( listener->object,
                    (guchar *)parseSSAPmsg_get_spaceid(msg),
                    (guchar *)parseSSAPmsg_get_nodeid(msg),
                    parseSSAPmsg_get_msgnumber(msg),
                    waiter->subscription_id) <0)
            {
           */
    if( m3_unsubscribe(sib_data, (gchar*) parseSSAPmsg_get_spaceid(msg),
                       (gchar*) parseSSAPmsg_get_nodeid(msg),
                       parseSSAPmsg_get_msgnumber(msg),
                       (gchar*) parseSSAPmsg_get_subscriptionid(msg)) < 0) {
        qDebug("Could not send unsubscribe request w/ msgid : %d\n", parseSSAPmsg_get_msgnumber( msg));

        status = ss_OperationFailed;

        bd = ssBufDesc_new();
        if( ss_StatusOK != ssBufDesc_CreateUnsubscribeResponse(bd,
                                                               (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                               (guchar *)parseSSAPmsg_get_spaceid(msg),
                                                               parseSSAPmsg_get_msgnumber(msg),
                                                               (msgStatus_t) status,
                                                               (ssElement_ct) parseSSAPmsg_get_subscriptionid(msg)))
        {
            ssBufDesc_free(&bd);
        }
    }
    else
    {
        rsp = nodemsgcontent_new_unsubscribe_rsp((guchar *)parseSSAPmsg_get_spaceid(msg),
                                                 (guchar *)parseSSAPmsg_get_nodeid(msg),
                                                 parseSSAPmsg_get_msgnumber(msg),
                                                 ss_StatusOK, (guchar*) parseSSAPmsg_get_subscriptionid(msg));
        if(rsp)
        {
            bd = ssBufDesc_new();
            if( ss_StatusOK != ssBufDesc_CreateUnsubscribeResponse(bd,
                                                                   (guchar *)parseSSAPmsg_get_nodeid(rsp),
                                                                   (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                                   parseSSAPmsg_get_msgnumber(rsp),
                                                                   parseSSAPmsg_get_msg_status(rsp),
                                                                   (guchar *)parseSSAPmsg_get_subscriptionid(rsp)))
            {
                ssBufDesc_free(&bd);
            }

            nodemsgcontent_free(&rsp);
        }
    }

    if(bd)
    {
        char* message = strdup((char*) ssBufDesc_GetMessage(bd));
        ssBufDesc_free(&bd);

        return message;
    }

    return NULL;
}

ssap_sib_message* m3_insert(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, EncodingType encoding, const char* insert_str)
{
    ssap_message_header *header;
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;
    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;
    ssStatus_t status = ss_StatusOK;
    //sib_op_parameter* param = (sib_op_parameter*) data;

    /* Allocate memory for message structs */
    header =  g_new0(ssap_message_header, 1);
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);
    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();
    rsp_msg->status = ss_StatusOK;    /* JUKKA - ARCES */


    header->space_id = spaceid;
    header->kp_id = nodeid;
    header->tr_id = msgnumber;
    req_msg->encoding = encoding;
    req_msg->insert_str = (gchar*)insert_str;

    qDebug("Parsed insert %d\n", header->tr_id);
    /* Initialize message structs */
    header->tr_type = M3_INSERT;
    header->msg_type = M3_REQUEST;


    if  (EncodingRDFXML == req_msg->encoding)
    {
        qDebug("INSERT: got insert msg encoding: RDFXML\n");
        rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");

    }
    /* Parse M3 triples here  */
    else if 	(EncodingM3XML == req_msg->encoding)
    {
        status = parseM3_triples_SIB(&(req_msg->insert_graph),
                                     req_msg->insert_str,
                                     NULL,
                                     &(rsp_msg->bnodes_str));
        // qDebug("INSERT: parsed bnodes str: %s\n", rsp_msg->bnodes_str);
        qDebug("Parsed M3 RDF in insert %d\n", header->tr_id);
        if (status != ss_StatusOK)
        {

            qWarning("INSERT: Parse failed, status code %d\n", status);
            rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");
            rsp_msg->status = status;
            goto send_response;
        }
    }


    s->header = header;
    s->req = req_msg;
    s->rsp = rsp_msg;
    s->op_lock = op_lock;
    s->op_cond = op_cond;
    s->op_complete = FALSE;

    g_async_queue_push(sib->insert_queue, s);

    /* Signal scheduler that new operation has been added to queue */
    g_mutex_lock(sib->new_reqs_lock);
    sib->new_reqs = TRUE;
    g_cond_signal(sib->new_reqs_cond);
    g_mutex_unlock(sib->new_reqs_lock);

    /* Block while operation is being processed */
    g_mutex_lock(s->op_lock);
    while (!(s->op_complete))
    {
        g_cond_wait(s->op_cond, s->op_lock);
    }
    s->op_complete = FALSE;
    g_mutex_unlock(op_lock);


    /* DAN - ARCES : XXX */
send_response:
    /* REMOVED rsp_msg->status = status; DAN - ARCES */

    /* REMOVE
        whiteboard_util_send_method_return(param->conn,
                param->msg,
                DBUS_TYPE_STRING, &(header->space_id),
                DBUS_TYPE_STRING, &(header->kp_id),
                DBUS_TYPE_INT32, &(header->tr_id),
                DBUS_TYPE_INT32, &(rsp_msg->status),
                DBUS_TYPE_STRING, &(rsp_msg->bnodes_str),
                WHITEBOARD_UTIL_LIST_END);
                */

    // WARNING: memory is not cleared
    if (EncodingM3XML == req_msg->encoding)
    {
        ssFreeTripleList(&(req_msg->insert_graph));
    }
    g_mutex_free(op_lock);
    g_cond_free(op_cond);
    g_free(s);

    g_free(req_msg);
    g_free(header);

    return rsp_msg;

    //g_free(rsp_msg->bnodes_str);
    //g_free(rsp_msg);
}

int m3_remove(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, EncodingType encoding, const char* remove_str)
{
    ssap_message_header *header;
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;
    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;
    ssStatus_t status = ss_StatusOK;
    //sib_op_parameter* param = (sib_op_parameter*) data;

    /* Allocate memory for message structs */
    header =  g_new0(ssap_message_header, 1);
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);
    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();
    rsp_msg->status = ss_StatusOK;    /* DAN - ARCES */

    /* REMOVE
    if(whiteboard_util_parse_message(param->msg,
            DBUS_TYPE_STRING, &(header->space_id),
            DBUS_TYPE_STRING, &(header->kp_id),
            DBUS_TYPE_INT32, &(header->tr_id),
            DBUS_TYPE_INT32, &(req_msg->encoding),
            DBUS_TYPE_STRING, &(req_msg->remove_str),
            DBUS_TYPE_INVALID) )
    {
    */

    header->space_id = spaceid;
    header->kp_id = nodeid;
    header->tr_id = msgnumber;
    req_msg->encoding = encoding;
    req_msg->remove_str = (gchar*) remove_str;

    // TODO: return error

    qDebug("Parsed remove %d\n", header->tr_id);
    /* Initialize message structs */
    header->tr_type = M3_REMOVE;
    header->msg_type = M3_REQUEST;
    if (EncodingRDFXML == req_msg->encoding)
    {
        qDebug("INSERT: got insert msg encoding: RDFXML\n");
        rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");

    }
    else if (EncodingM3XML == req_msg->encoding)
    {
        status = parseM3_triples(&(req_msg->remove_graph),
                                 req_msg->remove_str,
                                 NULL);
        if (status != ss_StatusOK)
        {
            qWarning("REMOVE: Parse failed, status code %d\n", status);
            rsp_msg->status = status; /* DAN - ARCES */
            goto send_response;
        }
    }
    s->header = header;
    s->req = req_msg;
    s->rsp = rsp_msg;
    s->op_lock = op_lock;
    s->op_cond = op_cond;
    s->op_complete = FALSE;

    g_async_queue_push(sib->insert_queue, s);

    /* Signal scheduler that new operation has been added to queue */
    g_mutex_lock(sib->new_reqs_lock);
    sib->new_reqs = TRUE;
    g_cond_signal(sib->new_reqs_cond);
    g_mutex_unlock(sib->new_reqs_lock);

    /* Block while operation is being processed */
    g_mutex_lock(s->op_lock);
    while (!(s->op_complete))
    {
        g_cond_wait(s->op_cond, s->op_lock);
    }
    s->op_complete = FALSE;
    g_mutex_unlock(op_lock);


send_response:
    /* REMOVE
        whiteboard_util_send_method_return(param->conn,
                param->msg,
                DBUS_TYPE_STRING, &(header->space_id),
                DBUS_TYPE_STRING, &(header->kp_id),
                DBUS_TYPE_INT32, &(header->tr_id),
                DBUS_TYPE_INT32, &(rsp_msg->status),
                WHITEBOARD_UTIL_LIST_END);
                */

    if (EncodingM3XML == req_msg->encoding)
    {
        ssFreeTripleList(&(req_msg->remove_graph));
    }

    g_mutex_free(op_lock);
    g_cond_free(op_cond);
    g_free(s);
    g_free(rsp_msg->bnodes_str);
    g_free(req_msg);
    g_free(rsp_msg);
    g_free(header);

    return status;
}

ssap_sib_message* m3_update(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, EncodingType encoding, char* insert_str, char* remove_str)
{
    ssap_message_header *header;
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;
    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;
    ssStatus_t status = ss_StatusOK;
    //sib_op_parameter* param = (sib_op_parameter*) data;

    /* Allocate memory for message structs */
    header =  g_new0(ssap_message_header, 1);
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);
    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();
    rsp_msg->status = ss_StatusOK;    /* DAN - ARCES */

    header->space_id = spaceid;
    header->kp_id = nodeid;
    header->tr_id = msgnumber;
    req_msg->encoding = encoding;
    req_msg->insert_str = insert_str;
    req_msg->remove_str = remove_str;

    /* Initialize message structs */
    header->tr_type = M3_UPDATE;
    header->msg_type = M3_REQUEST;
    /* Parse insert and remove strings to lists of triples*/

    if (EncodingRDFXML == req_msg->encoding)
    {
        qDebug("UPDATE: got udpate msg encoding: RDFXML\n");
        rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");

    }
    else if (EncodingM3XML == req_msg->encoding)
    {
        //Remove Part
        status = parseM3_triples(&(req_msg->remove_graph),
                                 req_msg->remove_str,
                                 NULL);
        if (status != ss_StatusOK)
        {
            qWarning("UPDATE: Remove parse failed, status code %d\n", status);
            rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");
            rsp_msg->status = status; /* DAN - ARCES */
            goto send_response;
        }



        status = parseM3_triples_SIB(&(req_msg->insert_graph),
                                     req_msg->insert_str,
                                     NULL,
                                     &(rsp_msg->bnodes_str));
        if (status != ss_StatusOK)
        {
            qWarning("UPDATE: Insert parse failed, status code %d\n", status);
            rsp_msg->bnodes_str = g_strdup("<urilist></urilist>");
            rsp_msg->status = status; /* DAN - ARCES */
            goto send_response;
        }
    }


    s->header = header;
    s->req = req_msg;
    s->rsp = rsp_msg;
    s->op_lock = op_lock;
    s->op_cond = op_cond;
    s->op_complete = FALSE;

    g_async_queue_push(sib->insert_queue, s);

    /* Signal scheduler that new operation has been added to queue */
    g_mutex_lock(sib->new_reqs_lock);
    sib->new_reqs = TRUE;
    g_cond_signal(sib->new_reqs_cond);
    g_mutex_unlock(sib->new_reqs_lock);

    /* Block while operation is being processed */
    g_mutex_lock(s->op_lock);
    while (!(s->op_complete))
    {
        g_cond_wait(s->op_cond, s->op_lock);
    }
    s->op_complete = FALSE;
    g_mutex_unlock(op_lock);


send_response:
    /* REMOVED rsp_msg->status = status; DAN - ARCES */

    if (EncodingM3XML == req_msg->encoding)
    {
        ssFreeTripleList(&(req_msg->remove_graph));
        ssFreeTripleList(&(req_msg->insert_graph));
    }

    g_mutex_free(op_lock);
    g_cond_free(op_cond);
    g_free(s);
    g_free(header);
    g_free(req_msg);

    //g_free(rsp_msg->bnodes_str);
    //g_free(rsp_msg);

    return rsp_msg;
}

ssap_sib_message* m3_query_spql_upd(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, query_type qtype, const char* query_str)
{
    ssap_message_header *header;
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;
    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;
    ssStatus_t status;
    //sib_op_parameter* param = (sib_op_parameter*) data;

    gboolean  is_sparql_update =FALSE;

    /* Allocate memory for message structs */
    header =  g_new0(ssap_message_header, 1);
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);
    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();

    /* REMOVE
    if(whiteboard_util_parse_message(param->msg,
            DBUS_TYPE_STRING, &(header->space_id),
            DBUS_TYPE_STRING, &(header->kp_id),
            DBUS_TYPE_INT32, &(header->tr_id),
            DBUS_TYPE_INT32, &(req_msg->type),
            DBUS_TYPE_STRING, &(req_msg->query_str),
            DBUS_TYPE_INVALID) )
    {
    */

    req_msg->type = qtype;
    req_msg->query_str = (gchar*) query_str;

    /* Initialize message structs */
    header->tr_type = M3_QUERY_SPQL_UPD;
    header->msg_type = M3_REQUEST;
    rsp_msg->status = ss_StatusOK;

    switch(qtype)
    {
    case QueryTypeTemplate:
        /* parsing triples from query_str to list */
        status = parseM3_triples(&(req_msg->template_query),
                                 (char*) query_str,
                                 NULL);
        break;
    case QueryTypeSPARQLSelect:
        //qDebug("SPARQL!\n");
        is_sparql_update=check_sparql_is_update((char*) query_str);
        status = ss_StatusOK; //FOR NOW!
        break;
    default: // Error //
        qWarning("UNREC!");
        rsp_msg->status = ss_SIBFailNotImpl;
        rsp_msg->results_str = g_strdup("");

        goto send_response;
        break;
    }

    if (! is_sparql_update)
    {

        if (status != ss_StatusOK)
        {
            if (status == ss_ParsingError)
                rsp_msg->status = ss_KPErrorMsgSyntax;
            else
                rsp_msg->status = status;

            rsp_msg->results_str = g_strdup("");
            goto send_response;
        }

        s->header = header;
        s->req = req_msg;
        s->rsp = rsp_msg;
        s->op_lock = op_lock;
        s->op_cond = op_cond;
        s->op_complete = FALSE;

        g_async_queue_push(sib->query_queue, s);

        /* Signal scheduler that new operation has been added to queue */
        g_mutex_lock(sib->new_reqs_lock);
        sib->new_reqs = TRUE;
        g_cond_signal(sib->new_reqs_cond);
        g_mutex_unlock(sib->new_reqs_lock);

        /* Block while operation is being processed */
        g_mutex_lock(s->op_lock);
        while (!(s->op_complete))
        {
            g_cond_wait(s->op_cond, s->op_lock);
        }
        s->op_complete = FALSE;
        g_mutex_unlock(op_lock);

        /* Generate results strings here */
        switch (qtype)
        {
        case QueryTypeTemplate:
            ssFreeTripleList(&(req_msg->template_query));
            rsp_msg->results_str = m3_gen_triple_string(rsp_msg->results);
            m3_free_triple_list_simple(&(rsp_msg->results));
            break;
        case QueryTypeSPARQLSelect:
            req_msg->template_query=NULL;
            break;
        default:
            qWarning("UNREC!");
            /* Error */
            /* Should never be reached */
            /* assert(0); */
            break;
        }
send_response:
        /* REMOVE
                whiteboard_util_send_method_return(param->conn,
                        param->msg,
                        DBUS_TYPE_STRING, &(header->space_id),
                        DBUS_TYPE_STRING, &(header->kp_id),
                        DBUS_TYPE_INT32, &(header->tr_id),
                        DBUS_TYPE_INT32, &(rsp_msg->status),
                        DBUS_TYPE_STRING, &(rsp_msg->results_str),
                        WHITEBOARD_UTIL_LIST_END);
                        */



        g_mutex_free(op_lock);
        g_cond_free(op_cond);
        g_free(s);
        g_free(req_msg);
        g_free(header);

        return rsp_msg;

        //g_free(rsp_msg->results_str);
        //g_free(rsp_msg);

        //return 0;
    }
    else
    {
        //SPARQL UPDATE
        s->header = header;
        s->req = req_msg;
        s->rsp = rsp_msg;
        s->op_lock = op_lock;
        s->op_cond = op_cond;
        s->op_complete = FALSE;

        ////////////////////////////////

        gboolean  insert_syntax_error;
        gboolean  remove_syntax_error;

        gboolean  insert_query_results;
        gboolean  remove_query_results;

        gboolean  delayed_update ;
        gboolean  delayed_update_outoftime ;

        gboolean series_update_followers = FALSE;

        gchar** split_update_array;
        gchar** pointer_split_update_array;

        gchar * temp_query= g_strdup(req_msg->query_str);
        split_update_array=g_strsplit( req_msg->query_str , sparql_seq_separator , 0);
        pointer_split_update_array = split_update_array;

        while (*pointer_split_update_array)
        {

            insert_syntax_error = FALSE;
            remove_syntax_error = FALSE;

            insert_query_results = FALSE;
            remove_query_results = FALSE;

            delayed_update = FALSE;
            delayed_update_outoftime = FALSE;

            gchar * remove_construct;
            //remove_construct= get_remove_construct(req_msg->query_str);
            remove_construct= get_remove_construct(*pointer_split_update_array);
            //if (remove_construct)
            //qDebug ("%s\n", remove_construct);


            gchar * insert_construct;
            insert_construct= get_insert_construct(*pointer_split_update_array);
            //if (insert_construct)
            //qDebug ("%s\n", insert_construct);


            if  (insert_construct || remove_construct)
            {

                double sparql_time_value =0;
                sparql_time_value = get_sparql_update_scheduled_time (*pointer_split_update_array);
                //sparql_time_value = get_sparql_update_scheduled_time (req_msg->query_str);

                struct timeval tv;
                gettimeofday(&tv, 0);
                double realtime= (int)tv.tv_sec + ((double)(int)tv.tv_usec / 1000000);

                if ((sparql_time_value > 0) && (sparql_time_value - realtime > 0))
                {
                    //DELAYED IN THE FUTURE
                    if (!series_update_followers)
                    {

                        rsp_msg->results_str=g_strdup(scheduler_upd_resp_true);
                        rsp_msg->status=ss_StatusOK;

                        /* REMOVE
                            whiteboard_util_send_method_return(param->conn,
                                    param->msg,
                                    DBUS_TYPE_STRING, &(header->space_id),
                                    DBUS_TYPE_STRING, &(header->kp_id),
                                    DBUS_TYPE_INT32, &(header->tr_id),
                                    DBUS_TYPE_INT32, &(rsp_msg->status),
                                    DBUS_TYPE_STRING, &(rsp_msg->results_str),
                                    WHITEBOARD_UTIL_LIST_END);
*/

                        g_free(rsp_msg->results_str);

                    }

                    usleep((int)((sparql_time_value - realtime)*1000000));
                    gboolean  delayed_update = TRUE;


                }
                else if ((sparql_time_value > 0) && (sparql_time_value - realtime < 0))
                {
                    //INTERVAL EXPIRED!  SENDING FALSE AND TERMINATING THE THREAD


                    //CHOICES:

                    //SENDING FALSE AND TERMINATING THE THREAD
                    /*
                         *
                         *
                        rsp_msg->results_str=g_strdup(sparql_upd_resp_false);
                        rsp_msg->status=ss_StatusOK;

                        whiteboard_util_send_method_return(param->conn,
                                param->msg,
                                DBUS_TYPE_STRING, &(header->space_id),
                                DBUS_TYPE_STRING, &(header->kp_id),
                                DBUS_TYPE_INT32, &(header->tr_id),
                                DBUS_TYPE_INT32, &(rsp_msg->status),
                                DBUS_TYPE_STRING, &(rsp_msg->results_str),
                                WHITEBOARD_UTIL_LIST_END);


                        g_free(rsp_msg->results_str);

                        g_free(remove_construct);
                        g_free(insert_construct);

                        g_mutex_free(op_lock);
                        g_cond_free(op_cond);

                        g_free(req_msg);
                        g_free(rsp_msg);
                        g_free(header);
                        g_free(s);

                        dbus_message_unref(param->msg);
                        g_free(param);

                        return NULL;
                        */


                    //  DO UPDATE EVEN IF OUT OF TIME ..
                    //  Do nothing and proceed normally.. Just sending false result
                    delayed_update_outoftime=TRUE;
                }
            }

            ///
            g_mutex_lock(sib->sparql_update_lock);
            ///

            m3_free_triple_list_simple(&(sib->RDF_list_insert));
            m3_free_triple_list_simple(&(sib->RDF_list_remove));

            if (remove_construct)
            {

                librdf_query * query;
                librdf_query_results * results;
                librdf_stream * stream;
                librdf_statement * statement;

                gchar * remove_construct_timed = convert_sib_time_calls( remove_construct);
                query=librdf_new_query(sib->RDF_world, "sparql", NULL, (unsigned char*) remove_construct_timed, NULL);
                results=librdf_model_query_execute(sib->RDF_model, query);
                g_free(remove_construct_timed);

                if (results == NULL)
                {
                    remove_syntax_error=TRUE;
                    //qDebug("SPARQL syntax error! \n");
                    librdf_free_query_results(results);
                    librdf_free_query(query);
                }
                else if (librdf_query_results_is_graph(results))
                {
                    stream=librdf_query_results_as_stream(results);
                    while(!librdf_stream_end(stream))
                    {
                        remove_query_results=TRUE;
                        sib->subs_sheduler_removed_triple=TRUE;

                        statement=librdf_stream_get_object(stream);
                        ////////////////
                        statement_in_tl(&(sib->RDF_list_remove), statement);
                        ////////////////
                        librdf_stream_next(stream);
                    }
                    librdf_free_stream(stream);

                }
                librdf_free_query_results(results);
                librdf_free_query(query);
            }
            if (insert_construct)
            {

                librdf_query * query;
                librdf_query_results * results;
                librdf_stream * stream;
                librdf_statement * statement;

                gchar * insert_construct_timed = convert_sib_time_calls( insert_construct);
                query=librdf_new_query(sib->RDF_world, "sparql", NULL, (unsigned char*) insert_construct_timed, NULL);
                results=librdf_model_query_execute(sib->RDF_model, query);
                g_free(insert_construct_timed);

                if (results == NULL)
                {
                    insert_syntax_error=TRUE;
                    //qDebug("SPARQL syntax error! \n");
                    librdf_free_query_results(results);
                    librdf_free_query(query);
                }
                else if (librdf_query_results_is_graph(results))
                {
                    stream=librdf_query_results_as_stream(results);

                    while(!librdf_stream_end(stream))
                    {
                        insert_query_results=TRUE;
                        sib->subs_sheduler_inserted_triple=TRUE;

                        statement=librdf_stream_get_object(stream);
                        ////////////////
                        statement_in_tl(&(sib->RDF_list_insert), statement);
                        ////////////////
                        librdf_stream_next(stream);
                    }
                    librdf_free_stream(stream);
                }
                librdf_free_query_results(results);
                librdf_free_query(query);

            }

            // UPDATING TS to REMOVE LIST

            GSList * pointer = sib->RDF_list_remove;
            librdf_statement * statement;

            while (pointer)
            {
                ssTriple_t * t = (ssTriple_t*) pointer->data;
                //Insert statement
                statement=librdf_new_statement(sib->RDF_world);

                librdf_statement_set_subject(statement,
                                             librdf_new_node_from_uri_string(sib->RDF_world,  t->subject));

                librdf_statement_set_predicate(statement,
                                               librdf_new_node_from_uri_string(sib->RDF_world, t->predicate));

                if (! t->objType) {
                    //URI: 	t->objType==0
                    librdf_statement_set_object(statement,
                                                librdf_new_node_from_uri_string(sib->RDF_world,  t->object));
                }
                else
                {	//Literal:  	t->objType==1
                    librdf_statement_set_object(statement,
                                                librdf_new_node_from_literal(sib->RDF_world, t->object, NULL, 0));
                }

                if ( sib->virtuoso )
                    librdf_model_context_remove_statement(sib->RDF_model, sib->virtuoso_context, statement);
                else
                    librdf_model_remove_statement(sib->RDF_model, statement);

                librdf_free_statement(statement);

                pointer=pointer->next;
            }

            // UPDATING TS to INSERT LIST

            pointer = sib->RDF_list_insert;
            while (pointer)
            {
                ssTriple_t * t = (ssTriple_t *) pointer->data;
                //Insert statement
                statement=librdf_new_statement(sib->RDF_world);

                librdf_statement_set_subject(statement,
                                             librdf_new_node_from_uri_string(sib->RDF_world,  t->subject));

                librdf_statement_set_predicate(statement,
                                               librdf_new_node_from_uri_string(sib->RDF_world, t->predicate));

                if (! t->objType) {
                    //URI: 	t->objType==0
                    librdf_statement_set_object(statement,
                                                librdf_new_node_from_uri_string(sib->RDF_world,  t->object));
                }
                else
                {
                    //Literal:  	t->objType==1
                    librdf_statement_set_object(statement,
                                                librdf_new_node_from_literal(sib->RDF_world, t->object, NULL, 0));
                }

                if ( sib->virtuoso)
                    librdf_model_context_add_statement(sib->RDF_model, sib->virtuoso_context, statement);
                else
                    librdf_model_add_statement(sib->RDF_model, statement);

                librdf_free_statement(statement);

                //Reasoning
                reasoning(sib, t, sib->enable_rdf_pp);

                pointer=pointer->next;
            }


            ////////////////////////////////////
            //   	IF TRIPLES INS / RM ->    //
            //	 		SCHEDULER EMULATOR	  //
            ////////////////////////////////////

            if (sib->RDF_list_insert || sib->RDF_list_remove)
            {

                //gpointer * op; //WARNING
                gpointer op;
                GSList * q_s_list=NULL;
                GHashTable* subs = sib->subs;
                GMutex* subscriptions_lock = sib->subscriptions_lock;

                g_async_queue_lock(sib->query_queue);
                while (NULL !=
                       (op = g_async_queue_try_pop_unlocked(sib->query_queue)))
                {
                    q_s_list = g_slist_prepend(q_s_list, op);
                    qDebug("Added item to query list");
                }
                g_async_queue_unlock(sib->query_queue);

                g_mutex_lock(subscriptions_lock);
                g_hash_table_foreach(subs, set_sub_to_pending, NULL);
                g_mutex_unlock(subscriptions_lock);

                g_slist_foreach(q_s_list, do_query_subscribe, sib);
                g_slist_free(q_s_list);
                q_s_list = NULL;

                /////////////////////////////////

            }
            g_mutex_unlock(sib->sparql_update_lock);



            if ((!delayed_update) && (!series_update_followers))
            {

                if  (
                     (insert_construct || remove_construct) &&
                     (!insert_syntax_error && !remove_syntax_error )
                     )
                {
                    if (delayed_update_outoftime)
                        rsp_msg->results_str=g_strdup(scheduler_upd_resp_false); //FOR NOW..
                    else
                    {
                        if (insert_query_results || remove_query_results)
                            rsp_msg->results_str=g_strdup(sparql_upd_resp_true);  //FOR NOW
                        else
                            rsp_msg->results_str=g_strdup(sparql_upd_resp_false); //FOR NOW
                    }
                    rsp_msg->status=ss_StatusOK;
                }
                else
                {
                    rsp_msg->results_str=g_strdup_printf("");
                    rsp_msg->status = ss_GeneralError;

                }

                /*
                        whiteboard_util_send_method_return(param->conn,
                                param->msg,
                                DBUS_TYPE_STRING, &(header->space_id),
                                DBUS_TYPE_STRING, &(header->kp_id),
                                DBUS_TYPE_INT32, &(header->tr_id),
                                DBUS_TYPE_INT32, &(rsp_msg->status),
                                DBUS_TYPE_STRING, &(rsp_msg->results_str),
                                WHITEBOARD_UTIL_LIST_END);
*/
                // TODO: WARNING! FIX WHILE CYCLE!!!!!!!

                // WARNING: temp memory clearing

                g_free(remove_construct);
                g_free(insert_construct);

                g_free(temp_query);
                g_strfreev (split_update_array);


                g_mutex_free(op_lock);
                g_cond_free(op_cond);

                g_free(req_msg);

                g_free(header);

                g_free(s);

                // END temp memory clearing

                return rsp_msg;

                g_free(rsp_msg->results_str);
                g_free(rsp_msg);

            }



            g_free(remove_construct);
            g_free(insert_construct);

            series_update_followers=TRUE;
            pointer_split_update_array ++;

        }

        g_free(temp_query);
        g_strfreev (split_update_array);


        g_mutex_free(op_lock);
        g_cond_free(op_cond);

        g_free(req_msg);
        g_free(rsp_msg);
        g_free(header);

        g_free(s);

    }
    //}
    /* REMOVE
    else
    {
        whiteboard_log_warning("Could not parse QUERY method call message\n");
    }
*/
    //dbus_message_unref(param->msg);
    //g_free(param);


    return NULL;
}

gpointer subscribe_thread(gpointer data)
{
    subscribe_param* param = (subscribe_param*) data;
    m3_subscribe(param->sib_data, (gchar*) param->spaceid, (gchar*) param->nodeid, param->msgnumber,
                 param->qtype, (const char*) param->query_str, &(param->subscription_id), &(param->results), param->sub_cond);
    GThread *thread = g_thread_self ();
    g_thread_unref(thread);

    g_thread_exit (NULL);
}


int m3_subscribe(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, query_type qtype, const char* query_str,
                 gchar** subscription_id, gchar** results, GCond* sub_cond)
{
    ssap_message_header *header;
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;

    ssStatus_t status;
    //sib_op_parameter* param = (sib_op_parameter*) data;

    gchar *space_id, *kp_id;
    gchar *temp_sub_id = NULL;

    gint tr_id;
    //member_data* kp_data;
    subscription_state* sub_state;

    /* Allocate memory for message structs */
    header =  g_new0(ssap_message_header, 1);
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);

    /* Initialize message structs */
    header->tr_type = M3_SUBSCRIBE;
    header->msg_type = M3_REQUEST;

    sub_state = g_new0(subscription_state, 1);

// WARNING! TODO: make check for valid parameters
/*
    if(whiteboard_util_parse_message(param->msg,
            DBUS_TYPE_STRING, &space_id,
            DBUS_TYPE_STRING, &kp_id,
            DBUS_TYPE_INT32, &tr_id,
            DBUS_TYPE_INT32, &(req_msg->type),
            DBUS_TYPE_STRING, &(req_msg->query_str),
            DBUS_TYPE_INVALID) )
    {
    */
    space_id = spaceid;
    kp_id = nodeid;
    tr_id = msgnumber;
    req_msg->type = qtype;
    req_msg->query_str = (gchar*) query_str;

        header->space_id = g_strdup(space_id);
        header->kp_id = g_strdup(kp_id);
        header->tr_id = tr_id;

        // TO DEFINE.. Commented because now just a waste of memory

        //g_mutex_lock(param->sib->members_lock);
        //kp_data = g_hash_table_lookup(param->sib->joined, kp_id);
        //if (kp_data != NULL)
        //{
        //  kp_data->subs = g_slist_prepend(kp_data->subs, g_strdup(temp_sub_id));
        //  kp_data->n_of_subs++;
        //}
        //g_mutex_unlock(param->sib->members_lock);

        /*
         * Initialize the query structure and start
           suitable subscription processor
         */

        switch(req_msg->type)
        {
        case QueryTypeTemplate:
            status = parseM3_triples(&(req_msg->template_query), req_msg->query_str,   NULL);
            rsp_msg->sparql_subscribe=FALSE;

            if (status == ss_StatusOK)
            {
                qDebug("status OK\n");
                /* ASSIGN SUB ID HERE *///////////////////////////////////////////
                temp_sub_id = g_strdup_printf("%s%s_%d",subs_uri, kp_id, tr_id);
qDebug("before mutex\n");
                g_mutex_lock(sib->subscriptions_lock);
qDebug("after mutex\n");

                while (g_hash_table_lookup(sib->subs, temp_sub_id)) {
                    /* Create better sub_id generator! */
                    g_free(temp_sub_id);
                    temp_sub_id = g_strdup_printf("%s%s_%d", subs_uri,kp_id, ++tr_id);
                }

                sub_state->status = M3_SUB_ONGOING;
                //sub_state->sub_id = g_strdup(temp_sub_id);
                //sub_state->kp_id = g_strdup(kp_id);

                g_hash_table_insert(sib->subs, temp_sub_id, sub_state);
                g_mutex_unlock(sib->subscriptions_lock);

                rsp_msg->sub_id = g_strdup(temp_sub_id);
                rsp_msg->first_subscribe =TRUE;

                //////////////////////////////////////////////////////
                rsp_msg->insert_remove_queue_async = g_async_queue_new();
                if (!rsp_msg->insert_remove_queue_async) exit(-1);

                ///////////////////////////////////////////////////////

                rsp_msg->sub_processing_on_thread_lock=g_mutex_new();
                if (!rsp_msg->sub_processing_on_thread_lock) exit(-1);

                generate_recycle_world_sub_ts(rsp_msg, sib);


                //qDebug("Started triples subscription with id %s", rsp_msg->sub_id); /* SUB_DEBUG */
                //starting

                status = (ssStatus_t) m3_subscribe_triples(header, req_msg, rsp_msg, sib, subscription_id, results, sub_cond);
                //ended

                g_async_queue_unref (rsp_msg->insert_remove_queue_async);

                g_mutex_free(rsp_msg->sub_processing_on_thread_lock);

                g_mutex_lock(sib->subscriptions_lock);
                g_hash_table_remove(sib->subs,temp_sub_id);
                g_mutex_unlock(sib->subscriptions_lock);

                g_free(temp_sub_id);

                destroy_recycle_world_sub_ts(rsp_msg, sib);

            }
            else
            {
                gchar *nullstr = g_strdup_printf("");
                /////////////////////////////////

                gchar tmp[37];
                uuid_t u1;
                uuid_generate(u1);
                uuid_unparse(u1, tmp);

                rsp_msg->sub_id = g_strdup_printf("%ssytax_error_id_/%s",subs_uri, tmp);

                //qDebug("SPARQL subscribe parsing wrong\n");
                /* REMOVE
                whiteboard_util_send_method_return(param->conn,
                        param->msg,
                        DBUS_TYPE_STRING, &(space_id),
                        DBUS_TYPE_STRING, &(kp_id),
                        DBUS_TYPE_INT32, &(tr_id),
                        DBUS_TYPE_INT32, &(status),
                        DBUS_TYPE_STRING, &(rsp_msg->sub_id),
                        DBUS_TYPE_STRING, &(nullstr),
                        WHITEBOARD_UTIL_LIST_END);
                        */


                *subscription_id = g_strdup(rsp_msg->sub_id);
                *results = g_strdup(nullstr);

                g_free(nullstr);

                g_cond_signal(sub_cond);

            }
            break;


        case QueryTypeSPARQLSelect:
            g_mutex_lock  (sib->sparql_preprocessing_lock);
            status = parseM3_sparql_bindings(&(req_msg->template_query), &(rsp_msg->sparql_vars),&(rsp_msg->required_sparql_vars), (unsigned char*) req_msg->query_str, sib);
            g_mutex_unlock(sib->sparql_preprocessing_lock);

            //qDebug("RECEIVED SPARQL IN\n");
            //qDebug("q: %s\n",req_msg->query_str);

            rsp_msg->sparql_query_str=g_strdup(req_msg->query_str);
            rsp_msg->sparql_subscribe=TRUE;
            /////////////////////////////////////////

            if (status == ss_StatusOK)
            {
                /* ASSIGN SUB ID HERE */
                temp_sub_id = g_strdup_printf("%s%s_%d",subs_uri, kp_id, tr_id);

                g_mutex_lock(sib->subscriptions_lock);
                while (g_hash_table_lookup(sib->subs, temp_sub_id)) {
                    /* Create better sub_id generator! */
                    g_free(temp_sub_id);
                    temp_sub_id = g_strdup_printf("%s%s_%d", subs_uri,kp_id, ++tr_id);
                }

                sub_state->status = M3_SUB_ONGOING;

                //sub_state->sub_id = g_strdup(temp_sub_id);
                //sub_state->kp_id = g_strdup(kp_id);

                g_hash_table_insert(sib->subs, temp_sub_id, sub_state);
                g_mutex_unlock(sib->subscriptions_lock);

                rsp_msg->sub_id = g_strdup(temp_sub_id);

                rsp_msg->first_subscribe =TRUE;

                // TODO: change exit(-1)
                rsp_msg->insert_remove_queue_async = g_async_queue_new();
                if (!rsp_msg->insert_remove_queue_async) exit(-1);

                rsp_msg->sub_processing_on_thread_lock=g_mutex_new();
                if (!rsp_msg->sub_processing_on_thread_lock) exit(-1);

                generate_recycle_world_sub_ts(rsp_msg, sib);

                //Starting thread loop
                status = (ssStatus_t) m3_subscribe_sparql(header, req_msg, rsp_msg, sib, subscription_id, results, sub_cond);
                //The thread loop is dead

                g_slist_free_full(rsp_msg->sparql_vars, g_free );
                g_slist_free_full(rsp_msg->required_sparql_vars, g_free );
                m3_additional_free_triple_sparql_list(&(req_msg->template_query));

                g_async_queue_unref (rsp_msg->insert_remove_queue_async);

                g_mutex_free(rsp_msg->sub_processing_on_thread_lock);

                g_mutex_lock(sib->subscriptions_lock);
                g_hash_table_remove(sib->subs,temp_sub_id);
                g_mutex_unlock(sib->subscriptions_lock);

                g_free(temp_sub_id);

                destroy_recycle_world_sub_ts(rsp_msg, sib);

            }
            else
            {
                gchar *nullstr = g_strdup_printf("");
                /////////////////////////////////

                gchar tmp[37];
                uuid_t u1;
                uuid_generate(u1);
                uuid_unparse(u1, tmp);


                rsp_msg->sub_id = g_strdup_printf("%ssytax_error_id_%s",subs_uri, tmp);
                //qDebug("SPARQL subscribe parsing wrong\n");
/* REMOVE
                whiteboard_util_send_method_return(param->conn,
                        param->msg,
                        DBUS_TYPE_STRING, &(space_id),
                        DBUS_TYPE_STRING, &(kp_id),
                        DBUS_TYPE_INT32, &(tr_id),
                        DBUS_TYPE_INT32, &(status),
                        DBUS_TYPE_STRING, &(rsp_msg->sub_id),
                        DBUS_TYPE_STRING, &(nullstr),
                        WHITEBOARD_UTIL_LIST_END);
                        */
                *subscription_id = g_strdup(rsp_msg->sub_id);
                *results = g_strdup(nullstr);

                g_free(nullstr);
                g_cond_signal(sub_cond);

            }

            break;
        }
        //qDebug("SUBSCRIBE: subscription %s finished \n", rsp_msg->sub_id);
    /* REMOVE
    }
    else
    {
        whiteboard_log_warning("Could not parse SUBSCRIBE method call message\n");
    }
    */

    // WARNING! TODO: memory clearing

    m3_free_triple_list_simple(&(req_msg->template_query));

    //g_free(req_msg->query_str);
    g_free(req_msg);

    g_free(rsp_msg->sub_id);
    g_free(rsp_msg);

    g_free(header->space_id);
    g_free(header->kp_id);
    g_free(header);

    //g_free(param);
// TODO: make return error values
    return 0;

}

gint m3_subscribe_triples(ssap_message_header *header,
        ssap_kp_message *req_msg,
        ssap_sib_message *rsp_msg,
        sib_data_structure* sib, gchar** subscription_id, gchar** results, GCond* sub_cond)
{

    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;

    gchar *space_id, *kp_id;
    gchar* sub_id = rsp_msg->sub_id;
    gint tr_id;
    subscription_state *sub_state;
    gboolean sub_ts_has_changed ;

    GMutex* thread_loop_lock;

    gboolean sub_processing_on_thread_local=TRUE;


    space_id = header->space_id;
    kp_id = header->kp_id;
    tr_id = header->tr_id;

    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();
    thread_loop_lock =  g_mutex_new();


    s->header = header;
    s->req = req_msg;
    s->rsp = rsp_msg;
    s->op_lock = op_lock;
    s->op_cond = op_cond;
    s->op_complete = FALSE;
    s->thread_loop_lock=thread_loop_lock;


    sib_op_parameter* p;
    p = g_new0(sib_op_parameter, 1);
    p->sib = sib;
    p->operation = M3_SUBSCRIBE;

    s->rsp->param = (gpointer *) p;

    g_async_queue_push(sib->query_queue, s);

    // Signal scheduler that new operation has been added to queue
    g_mutex_lock(sib->new_reqs_lock);

    sib->new_reqs = TRUE;
    // qDebug("Now signaling SCHEDULER in SUBSCRIBE tr %d\n", header->tr_id);
    g_cond_signal(sib->new_reqs_cond);
    g_mutex_unlock(sib->new_reqs_lock);

    // Block while operation is being processed
    g_mutex_lock(s->op_lock);
    while (!(s->op_complete))
    {
        g_cond_wait(s->op_cond, s->op_lock);
    }
    s->op_complete = FALSE;
    g_mutex_unlock(op_lock);
    //qDebug("Got baseline query result for subscription %s\n", rsp_msg->sub_id);

    rsp_msg->results_str = m3_gen_triple_string(rsp_msg->results);
    m3_free_triple_list_simple(&(rsp_msg->results));

    /*
    whiteboard_util_send_method_return(param->conn,
            param->msg,
            DBUS_TYPE_STRING, &(space_id),
            DBUS_TYPE_STRING, &(kp_id),
            DBUS_TYPE_INT32, &(tr_id),
            DBUS_TYPE_INT32, &(rsp_msg->status),
            DBUS_TYPE_STRING, &(rsp_msg->sub_id),
            DBUS_TYPE_STRING, &(rsp_msg->results_str),
            WHITEBOARD_UTIL_LIST_END);
            */

    *subscription_id = g_strdup(rsp_msg->sub_id);
    *results = g_strdup(rsp_msg->results_str);

    g_cond_signal(sub_cond);

    g_free(rsp_msg->results_str);
    //dbus_message_unref(param->msg);

    rsp_msg->results=NULL;

    do
    {

        //qDebug("Loop sub thread %s  \n", sub_id);

        // Check if unsubscribed //
        g_mutex_lock(sib->subscriptions_lock);
        sub_state = (subscription_state*)g_hash_table_lookup(sib->subs, sub_id);
        g_mutex_unlock(sib->subscriptions_lock);

        if (sub_state->status == M3_SUB_STOPPED)
        {
            g_mutex_lock(sib->subscriptions_lock);

            g_free(sub_state->sub_id);
            g_free(sub_state->kp_id);

            g_hash_table_remove(sib->subs, sub_id);
            sub_state->unsub = TRUE;

            g_cond_signal(sub_state->unsub_cond);
            // UNLOCK SUBSCRIPTION TABLE /
            g_mutex_unlock(sib->subscriptions_lock);
            // UNLOCK UNSUB LOCK //

            g_free(rsp_msg->sparql_query_str);

            g_mutex_free(thread_loop_lock);
            g_cond_free(op_cond);
            g_mutex_free(op_lock);
            g_free(s);

            break;
        }

        g_mutex_lock(sib->subscriptions_lock);
        sub_state = (subscription_state*)g_hash_table_lookup(sib->subs, sub_id);
        g_mutex_unlock(sib->subscriptions_lock);

        if (sub_state->status == M3_SUB_PENDING)
        {
            g_mutex_lock(sib->new_reqs_lock);
            sib->new_reqs = TRUE;
            g_cond_signal(sib->new_reqs_cond);
            g_mutex_unlock(sib->new_reqs_lock);
            //qDebug("Subscription %s pending, setting new_reqs flag\n", rsp_msg->sub_id); // SUB_DEBUG //
        }

        // Block while operation is processed //
        g_mutex_lock(s->thread_loop_lock);
        s->thread_busy=FALSE;
        g_mutex_unlock(s->thread_loop_lock);

        //g_cond_signal(s->thread_is_free);

        g_mutex_lock(s->op_lock);
        while (!(s->op_complete))
        {
            qDebug("WAIT\n");
            g_cond_wait(s->op_cond, s->op_lock);
            qDebug("AFTER WAIT\n");
        }
        s->op_complete = FALSE;
        g_mutex_unlock(op_lock);

        g_mutex_lock(s->thread_loop_lock);
        s->thread_busy=TRUE;
        g_mutex_unlock(s->thread_loop_lock);

        qDebug("Got new subscription result for transaction %d\n", tr_id);
        //qDebug("Got new query result for subscription %s\n", rsp_msg->sub_id); //

        g_mutex_lock(rsp_msg->sub_processing_on_thread_lock);
        sub_processing_on_thread_local=rsp_msg->sub_processing_on_thread;
        g_mutex_unlock(rsp_msg->sub_processing_on_thread_lock);

        qDebug("BEFORE SUBPROCESSING\n");

        if (sub_processing_on_thread_local)
        {

            ////////////////////
            g_mutex_lock(sib->sub_threads_lock);
            //qDebug("full active th: %d ", param->sib->sub_threads_active);
            if (sib->sub_threads_active  <  sib->sub_threads_max)
                sib->sub_threads_active ++;
            g_mutex_unlock(sib->sub_threads_lock);
            ////////////////////

            gboolean result;
            qDebug("LOAD SUBSCRIPTIONS\n");
            result= load_rdf_subscriptions(s, sib);
            //qDebug("Loaded on thread, results found : %d\n", result);
            qDebug("LOADED SUBSCRIPTIONS\n");
            ////////////////////
            g_mutex_lock(sib->sub_threads_lock);
            if (sib->sub_threads_active  > 0)
                sib->sub_threads_active --;
            g_mutex_unlock(sib->sub_threads_lock);
            ////////////////////

        }	//END FULL THREAD CASE


    }
    while(TRUE);

    return ss_StatusOK;
}


gint m3_subscribe_sparql(ssap_message_header *header,ssap_kp_message *req_msg, ssap_sib_message *rsp_msg,
                         sib_data_structure* sib, gchar** subscription_id, gchar** results, GCond* sub_cond){

    GMutex* op_lock;
    GCond* op_cond;
    scheduler_item* s;

    gchar *space_id, *kp_id;
    gchar* sub_id = rsp_msg->sub_id;
    gint tr_id;

    subscription_state *sub_state;

    gboolean sub_processing_on_thread_local=TRUE;

    GMutex* thread_loop_lock;

    space_id = header->space_id;
    kp_id = header->kp_id;
    tr_id = header->tr_id;

    s = g_new0(scheduler_item, 1);
    op_lock = g_mutex_new();
    op_cond = g_cond_new();

    thread_loop_lock =  g_mutex_new();

    s->header = header;
    s->req = req_msg;
    s->rsp = rsp_msg;
    s->op_lock = op_lock;
    s->op_cond = op_cond;
    s->op_complete = FALSE;

    s->thread_loop_lock=thread_loop_lock;

    s->thread_is_free=g_cond_new();
    s->thread_is_free_lock=g_mutex_new();

    //s->rsp->param= param; WARNING

    //Time utilities
    struct timeval start, end;
    unsigned long mtime, seconds, useconds, global_t =0;

    int number_of_items;

    //rsp_msg->sparql_query_str=g_strdup(req_msg->query_str);

    g_async_queue_push(sib->query_queue, s);

    /* Signal scheduler that new operation has been added to queue */
    g_mutex_lock(sib->new_reqs_lock);
    sib->new_reqs = TRUE;
    // qDebug("Now signaling SCHEDULER in SUBSCRIBE tr %d\n", header->tr_id);
    g_cond_signal(sib->new_reqs_cond);
    g_mutex_unlock(sib->new_reqs_lock);

    /* Block while operation is being processed */
    g_mutex_lock(s->op_lock);
    while (!(s->op_complete))
    {
        g_cond_wait(s->op_cond, s->op_lock);
    }
    s->op_complete = FALSE;
    g_mutex_unlock(s->op_lock);


    if (sib->debug_latency)
    {	//TIME MEASUREMENT UTILITY LINES
        gettimeofday(&start, NULL);
        /////////////////////////////
    }

    //////////////////////////////////////
    qDebug("SPARQL first execution on the created context\n");
    rsp_msg->results_str= (gchar*) load_sparql_on_dedicate_sub_ts((unsigned char*) s->rsp->sparql_query_str,s->rsp->subworld,s->rsp->substorage, s->rsp->submodel);
    //////////////////////////////////////
/*
    whiteboard_util_send_method_return(param->conn,
            param->msg,
            DBUS_TYPE_STRING, &(space_id),
            DBUS_TYPE_STRING, &(kp_id),
            DBUS_TYPE_INT32, &(tr_id),
            DBUS_TYPE_INT32, &(rsp_msg->status),
            DBUS_TYPE_STRING, &(rsp_msg->sub_id),
            DBUS_TYPE_STRING, &(rsp_msg->results_str),
            WHITEBOARD_UTIL_LIST_END);

    dbus_message_unref(param->msg);
*/
    *subscription_id = g_strdup(rsp_msg->sub_id);
    *results = g_strdup(rsp_msg->results_str);

    g_cond_signal(sub_cond);

    g_free(rsp_msg->results_str);

    if (sib->debug_latency)
    {
        ///////////////////////////////////////
        gettimeofday(&end, NULL);
        seconds  = end.tv_sec  - start.tv_sec;
        useconds = end.tv_usec - start.tv_usec;
        fprintf(stderr,"Thread First Sparql Time [us,sub_id] : %ld , %s \n",  seconds * 1000000 + useconds, rsp_msg->sub_id);
        ///////////////////////////////////////
    }

    do
    {
        //qDebug("Loop sub thread %s	\n", sub_id);

        /* Check if unsubscribed */
        g_mutex_lock(sib->subscriptions_lock);
        sub_state = (subscription_state*)g_hash_table_lookup(sib->subs, sub_id);

        if (sub_state->status == M3_SUB_STOPPED)
        {

            g_free(sub_state->sub_id);
            g_free(sub_state->kp_id);

            g_hash_table_remove(sib->subs, sub_id);
            sub_state->unsub = TRUE;

            g_cond_signal(sub_state->unsub_cond);
            /* UNLOCK SUBSCRIPTION TABLE */
            g_mutex_unlock(sib->subscriptions_lock);
            /* UNLOCK UNSUB LOCK */

            g_free(rsp_msg->sparql_query_str);

            g_mutex_free(thread_loop_lock);
            g_cond_free(op_cond);
            g_mutex_free(op_lock);
            g_free(s);

            break;
        }

        sub_state = (subscription_state*)g_hash_table_lookup(sib->subs, sub_id);

        if (sub_state->status == M3_SUB_PENDING)
        {
            g_mutex_unlock(sib->subscriptions_lock);

            g_mutex_lock(sib->new_reqs_lock);
            sib->new_reqs = TRUE;
            g_cond_signal(sib->new_reqs_cond);
            g_mutex_unlock(sib->new_reqs_lock);
            //qDebug("Subscription %s pending, setting new_reqs flag\n", rsp_msg->sub_id); /* SUB_DEBUG */
        }
        else
        {
            g_mutex_unlock(sib->subscriptions_lock);
        }


        /* Block while operation is processed */
        g_mutex_lock(s->thread_loop_lock);
        s->thread_busy=FALSE;
        g_mutex_unlock(s->thread_loop_lock);

        //g_cond_signal(s->thread_is_free);

        g_mutex_lock(s->op_lock);
        while (!(s->op_complete))
        {
            g_cond_wait(s->op_cond, s->op_lock);
        }
        s->op_complete = FALSE;
        g_mutex_unlock(s->op_lock);

        if (sib->debug_latency)
        {
            //TIME MEASUREMENT UTILITY LINES
            ////////////////////////
            gettimeofday(&start, NULL);
            /////////////////////////////
        }

        g_mutex_lock(s->thread_loop_lock);
        s->thread_busy=TRUE;
        g_mutex_unlock(s->thread_loop_lock);

        qDebug("Got new subscription result for transaction %d\n", tr_id);
        //qDebug("Got new query result for subscription %s\n", rsp_msg->sub_id); /* SUB_DEBUG */

        g_mutex_lock(rsp_msg->sub_processing_on_thread_lock);
        sub_processing_on_thread_local=rsp_msg->sub_processing_on_thread;
        g_mutex_unlock(rsp_msg->sub_processing_on_thread_lock);

        if (sub_processing_on_thread_local)
        {

            ////////////////////
            g_mutex_lock(sib->sub_threads_lock);
            //qDebug("full active th: %d ", param->sib->sub_threads_active);
            if (sib->sub_threads_active  <  sib->sub_threads_max)
                sib->sub_threads_active ++;
            g_mutex_unlock(sib->sub_threads_lock);
            ////////////////////

            gboolean result;
            result = load_sparql_booster_subscriptions(s, sib);
            //qDebug ("Loaded SPARQL on thread, results found : %d\n", result);

            ////////////////////
            g_mutex_lock(sib->sub_threads_lock);
            if (sib->sub_threads_active  > 0)
                sib->sub_threads_active --;
            g_mutex_unlock(sib->sub_threads_lock);
            ////////////////////

        }	//END FULL THREAD CASE

        if (sib->debug_latency)
        {
            ///////////////////////////////////////
            gettimeofday(&end, NULL);
            seconds  = end.tv_sec  - start.tv_sec;
            useconds = end.tv_usec - start.tv_usec;
            global_t = seconds * 1000000 + useconds;
            fprintf(stderr,"Thread Sparql time [us,sub_id] : %d , %s \n",  global_t , rsp_msg->sub_id);
            ///////////////////////////////////////
        }
    }
    while(TRUE);

    return ss_StatusOK;
}

int m3_unsubscribe(sib_data_structure* sib, gchar* spaceid, gchar* nodeid, gint msgnumber, gchar* sub_id)
{
    ssap_kp_message *req_msg;
    ssap_sib_message *rsp_msg;

    GCond* unsub_cond;
    subscription_state* sub;

    unsub_cond = g_cond_new();
    /* Allocate memory for message structs */
    req_msg = g_new0(ssap_kp_message, 1);
    rsp_msg = g_new0(ssap_sib_message, 1);
/*
    if(whiteboard_util_parse_message(param->msg,
            DBUS_TYPE_STRING, &(header->space_id),
            DBUS_TYPE_STRING, &(header->kp_id),
            DBUS_TYPE_INT32, &(header->tr_id),
            DBUS_TYPE_STRING, &(req_msg->sub_id),
            DBUS_TYPE_INVALID) )
    {
    */
    req_msg->sub_id = sub_id;

        g_mutex_lock(sib->subscriptions_lock);
        sub = (subscription_state*)g_hash_table_lookup(sib->subs, req_msg->sub_id);
        g_mutex_unlock(sib->subscriptions_lock);

        if (NULL != sub)
        {
            qDebug("UNSUBSCRIBE: Found sub for sub id %s\n", req_msg->sub_id);

            /* Set subscription status to stopped */
            g_mutex_lock(sib->subscriptions_lock);
            sub->unsub_cond = unsub_cond;
            sub->unsub = FALSE;
            sub->status = M3_SUB_STOPPED;
            g_mutex_unlock(sib->subscriptions_lock);

            /* Signal scheduler to execute a round to get
             all subscriptions to check their status
             TODO: better way to signal subscriptions
             */
            g_mutex_lock(sib->new_reqs_lock);
            sib->new_reqs = TRUE;
            g_cond_signal(sib->new_reqs_cond);
            g_mutex_unlock(sib->new_reqs_lock);

            /* Wait until subscription processing has finished */

            g_mutex_lock(sib->subscriptions_lock);
            while(!sub->unsub)
            {
                g_cond_wait(unsub_cond, sib->subscriptions_lock);
            }
            sub->unsub = FALSE;
            //g_free(sub);

            g_mutex_unlock(sib->subscriptions_lock);

            rsp_msg->status = ss_StatusOK;
        }
        else
            rsp_msg->status = ss_KPErrorRequest;

        /*
        whiteboard_util_send_method_return(param->conn,
                param->msg,
                DBUS_TYPE_STRING, &(header->space_id),
                DBUS_TYPE_STRING, &(header->kp_id),
                DBUS_TYPE_INT32, &(header->tr_id),
                DBUS_TYPE_INT32, &(rsp_msg->status),
                DBUS_TYPE_STRING, &(req_msg->sub_id),
                WHITEBOARD_UTIL_LIST_END);*/

        qDebug("UNSUBSCRIBE: Sent unsub cnf for sub id %s\n", req_msg->sub_id);
/*
    }
    else
    {
        whiteboard_log_warning("Could not parse UNSUBSCRIBE method call message\n");
    }
    */

    g_free(req_msg);
    g_free(rsp_msg);

    g_cond_free(unsub_cond);

    return 0;
}

sib_data_structure* sib_initialize(gchar* name, gpointer parameter_list)
{
    sib_data_structure* sd;

    GSList* param_list = (GSList*) parameter_list;
    /* Allocate sib data structures */

    sd = g_new0(sib_data_structure, 1);
    if (NULL == sd) exit(-1);

    /* Initialize smart space name */
    sprintf(sd->ss_name, "%s-%s", name, getlogin());

    sd->insert_queue = g_async_queue_new();
    if (NULL == sd->insert_queue) exit(-1);

    sd->query_queue = g_async_queue_new();
    if (NULL == sd->query_queue) exit(-1);

    sd->used_worlds_async_queue= g_async_queue_new();
    if (NULL == sd->used_worlds_async_queue) exit(-1);

    sd->joined = g_hash_table_new(g_str_hash, g_str_equal);
    if (NULL == sd->joined) exit(-1);

    sd->subs = g_hash_table_new(g_str_hash, g_str_equal);
    if (NULL == sd->subs) exit(-1);

    //sd->members_lock = g_mutex_new();
    //if (NULL == sd->members_lock) exit(-1);

    sd->subscriptions_lock = g_mutex_new();
    if (NULL == sd->subscriptions_lock) exit(-1);

    sd->new_reqs_lock = g_mutex_new();
    if (NULL == sd->new_reqs_lock) exit(-1);

    sd->new_reqs_cond = g_cond_new();
    if (NULL == sd->new_reqs_cond) exit(-1);

    sd->dbus_indication_lock=g_mutex_new();
    if (!sd->dbus_indication_lock) exit(-1);

    sd->sub_threads_lock=g_mutex_new();
    if (NULL == sd->sub_threads_lock) exit(-1);

    sd->sparql_update_lock=g_mutex_new();
    if (NULL == sd->sparql_update_lock) exit(-1);

    sd->new_reqs = FALSE;

    //INITIALIZE REDLAND ENVIROMENT
    sd->RDF_world=librdf_new_world();
    librdf_world_open(sd->RDF_world);
    rasqal_world_set_warning_level(librdf_world_get_rasqal(sd->RDF_world), 0);

    //CHECK FOR PARAMETERS
    sd->enable_rdf_pp   =   FALSE;
    sd->enable_protection =    FALSE;
    sd->mem_volatile      =     FALSE;
    sd->sqlite =        FALSE;
    sd->virtuoso =  FALSE;
    sd->subs_persistent   =  FALSE;

    sd->debug_latency = FALSE;

    //////////INITIALIZED TO SEQUENTIAL SUBSCRIBE////////
    sd->sub_threads_max=INT_MAX;
    sd->sub_threads_active=0;
    /////////////////////////////////////////////////////

    for ( ; param_list != NULL ; param_list = g_slist_next(param_list))
    {
        gchar* current_parameter= (gchar*) param_list->data;
        if (strcmp ((char*)"--enable-protections",(char*)current_parameter)==0)
        {
            qDebug  ("Launching with write protections option\n");
            sd->enable_protection = FALSE;
        }
        if ((strcmp ((char*)"--storage-volatile",(char*)current_parameter)==0) || (strcmp ((char*)"--ram",(char*)current_parameter)==0))
        {
            qDebug  ("Launching with no persistent DB\n");
            sd->mem_volatile = TRUE;
            sd->mem_volatile_hash = FALSE;
        } else if
            ((strcmp ((char*)"--storage-volatile-hash",(char*)current_parameter)==0) || (strcmp ((char*)"--ram-hash",(char*)current_parameter)==0))
        {
            qDebug  ("Launching with non persistent DB using Hash\n");
            sd->mem_volatile = FALSE;
            sd->mem_volatile_hash = TRUE;
        }
        if (strcmp ((char*)"--storage-sqlite",(char*)current_parameter)==0)
        {
            qDebug  ("Launching with Sqlite storage (SLOW!)\n");
            sd->sqlite = TRUE;
        }

        /*if (strcmp ((char*)"--storage-virtuoso",(char*)current_parameter)==0)
        {

            qDebug  ("Loading Virtuoso, working on graph: <http://sib#%s> \n", sd->ss_name);
            sd->virtuoso = TRUE;

            qDebug  ( "\n");
            qDebug  ( "Setting parameters..      \n");
            char host [100] ;
            qDebug  ( "\n");
            qDebug               ("host (e.g. =localhost): ");
            scanf("%s", &host);
            char user [50];
            qDebug               ("user (e.g. =dba)     : ");
            scanf("%s", &user);
            char *pass = getpass("password             : ");
            qDebug  ("\n");
            sd->virtuoso_param=g_strdup_printf("dsn='VOS', host='%s', user='%s',password='%s'",host,user,pass);
            //qDebug  ("%s\n",sd->virtuoso_param);

        }*/
        /*if (strstr (current_parameter,(char*)"--storage-virtuoso-p" ) != NULL)
        {

            qDebug  ("Loading Virtuoso, working on graph: <http://sib#%s> \n", sd->ss_name);
            sd->virtuoso = TRUE;

            gchar** split_param_point =NULL;
            gchar** split_param =NULL;
            gchar * params;

            char * tempparam = strdup(current_parameter);
            split_param=g_strsplit(tempparam, "==" , 0);
            split_param_point=split_param;
            split_param++;

            params = string_substitution_(*split_param,"\"", "");

            sd->virtuoso_param=g_strdup(params);

        }*/
        if (strcmp ((char*)"--subs-persistent",(char*)current_parameter)==0)
        {
            qDebug  ("Launching with BDB persistent subscriptions triplestores\n");
            sd->subs_persistent = TRUE;
            sd->subs_hash=FALSE;
        }
        else if (strcmp ((char*)"--subs-hash",(char*)current_parameter)==0)
        {
            qDebug  ("Launching with Hash volatile subscriptions triplestores\n");
            sd->subs_persistent = FALSE;
            sd->subs_hash=TRUE;
        }

        if (strcmp ((char*)"--enable-rdf++",(char*)current_parameter)==0)
        {
            qDebug  ("Launching Reasoner. No persistent DB will be rescued !\n");
            sd->enable_rdf_pp = TRUE;
        }
        if (strstr (current_parameter,(char*)"--threads" ) != NULL)
        {

            gchar** split_param_point =NULL;
            gchar** split_param =NULL;
            char * tempparam = strdup(current_parameter);
            split_param=g_strsplit(tempparam, "=" , 0);
            split_param_point=split_param;
            split_param++;

            if (g_strcmp0(*split_param,"inf")==0)
            {
                sd->sub_threads_max=INT_MAX;
                qDebug  ("Setting number of sparql booster threads to unlimited\n", sd->sub_threads_max);
            }
            else
            {
                sd->sub_threads_max=atoi(*split_param);
                if (sd->sub_threads_max == NULL)
                {
                    qDebug ( "Setting no sparql booster threads\n\n");
                }
                else
                    qDebug ( "Setting number of sparql booster threads to : %d\n", sd->sub_threads_max);
            }
            g_strfreev (split_param_point);

        }

        if (strstr (current_parameter,(char*)"--debug-latency" ) != NULL)
        {
            qDebug ( "Setting debug latency\n\n");
            sd->debug_latency=TRUE;
        }

    }

    //////Creating Model and storage ////////////////////////
    initialize_storage_model(sd);
    ////////////////////////////////////////////////////////

    //ARCES DIFFERENTIAL SUBS INS/REM SCHEDULER TRIPLESTORES//////////////////////////////////////////////////////////
    sd->RDF_list_insert=NULL;
    sd->RDF_list_remove=NULL;

    sd->temp_ins_rem_operations_lock = g_mutex_new();

    //sparql subscribe modules
    sd->sparql_preprocessing_world = rasqal_new_world();
    rasqal_world_set_warning_level(sd->sparql_preprocessing_world, 0);
    sd->sparql_preprocessing_lock = g_mutex_new();



    ////REASONING INIT
    sd->subClasses =    g_hash_table_new_full(g_str_hash, g_str_equal,  g_free, NULL);
    sd->subProperties =     g_hash_table_new_full(g_str_hash, g_str_equal,  g_free, NULL);
    sd->propertyDomain=     g_hash_table_new_full(g_str_hash, g_str_equal,  g_free, NULL);
    sd->propertyRange=  g_hash_table_new_full(g_str_hash, g_str_equal,  g_free, NULL);
    ////--REASONING

    /* Start scheduler */
    g_thread_create(scheduler, sd, FALSE, NULL);


    if (NULL == sd->new_reqs_cond) exit(-1);
    if (NULL == sd->RDF_model) exit(-1);

    return sd;
}

gpointer scheduler(gpointer data)
{
    sib_data_structure* p = (sib_data_structure*)data;

    GAsyncQueue* i_queue = p->insert_queue;
    GAsyncQueue* q_queue = p->query_queue;

    GCond* new_reqs_cond = p->new_reqs_cond;
    GMutex* new_reqs_lock = p->new_reqs_lock;

    GHashTable* subs = p->subs;
    GMutex* subscriptions_lock = p->subscriptions_lock;

    gboolean updated = false;

    GSList* i_list = NULL;
    GSList* q_s_list = NULL;

    scheduler_item* op;

    //time managment
    struct timeval start, end;
    unsigned long mtime, seconds, useconds;


    while (TRUE)
    {
        /* Wait until there is actually something in the queues */
        g_mutex_lock(new_reqs_lock);
        while (!(p->new_reqs))
        {
            g_cond_wait(new_reqs_cond, new_reqs_lock);
        }
        p->new_reqs = FALSE;
        g_mutex_unlock(new_reqs_lock);

        //Freeze scheduler to perform sparql updates
        g_mutex_lock(p->sparql_update_lock);

        qDebug("Schedule Cycle Begin\n");

        /* Lock insert and query queues
        Insert contents into lists for processing
        */

        g_async_queue_lock(i_queue);
        while (NULL !=
                (op = (scheduler_item*)g_async_queue_try_pop_unlocked(i_queue)))
        {

            if (p->enable_protection)
            {
                /*AD-ARCES*/
                qDebug("#######################################\n");
                qDebug("PROTECTION CONTROL \n");
                qDebug("#######################################\n");

                qWarning("Not implemented!\n");
                //ProtectionCompatibilityFilter(op);
            }

            i_list = g_slist_prepend(i_list, op);
            qDebug("Added item to insert list");
        }
        g_async_queue_unlock(i_queue);

        if (i_list != NULL)
            updated = true;

        g_async_queue_lock(q_queue);
        while (NULL !=
                (op = (scheduler_item*)g_async_queue_try_pop_unlocked(q_queue)))
        {
            q_s_list = g_slist_prepend(q_s_list, op);
            qDebug("Added item to query list");
        }
        g_async_queue_unlock(q_queue);

        //Clean Redland Temp Ins/Rem Triplestores
        //g_mutex_lock(p->temp_ins_rem_operations_lock);
        m3_free_triple_list_simple(&(p->RDF_list_insert));
        m3_free_triple_list_simple(&(p->RDF_list_remove));
        //g_mutex_unlock(p->temp_ins_rem_operations_lock);

        p->subs_sheduler_inserted_triple= FALSE;
        p->subs_sheduler_removed_triple=  FALSE;


        if (p->debug_latency && i_list)
        {
            //TIME MEASUREMENT UTILITY LINES
            //////////////////////////
            gettimeofday(&start, NULL);
            /////////////////////////////
        }

        g_slist_foreach(i_list, do_insert, p);
        ///////////////////////////////////////
        if (p->debug_latency && i_list)
        {
            gettimeofday(&end, NULL);
            seconds  = end.tv_sec  - start.tv_sec;
            useconds = end.tv_usec - start.tv_usec;
            fprintf(stderr,"Scheduler Insert T [us] :%d \n", (seconds*1000000 + useconds));
        }
        ///////////////////////////////////////
        g_slist_free(i_list);
        i_list = NULL;



        if (updated)
        {
            g_mutex_lock(subscriptions_lock);
            g_hash_table_foreach(subs, set_sub_to_pending, NULL);
            g_mutex_unlock(subscriptions_lock);

            updated = false;
        }

        if (p->debug_latency && q_s_list)
        {
            //TIME MEASUREMENT UTILITY LINES
            //////////////////////////
            gettimeofday(&start, NULL);
            /////////////////////////////
        }
        //qDebug("RDF store updated, set all subscriptions to pending\n"); /* SUB_DEBUG */
        g_slist_foreach(q_s_list, do_query_subscribe, p);
        ///////////////////////////////////////
        if (p->debug_latency && q_s_list)
        {
            gettimeofday(&end, NULL);
            seconds  = end.tv_sec  - start.tv_sec;
            useconds = end.tv_usec - start.tv_usec;
            fprintf(stderr,"Scheduler Query T [us] : %d \n", (seconds*1000000 + useconds));
        }
        ///////////////////////////////////////
        g_slist_free(q_s_list);
        q_s_list = NULL;

        //End freeze scheduler to perform sparql updates
        g_mutex_unlock(p->sparql_update_lock);

    }

}

void do_insert(gpointer op_param, gpointer p_param)
{
    scheduler_item* op = (scheduler_item*) op_param;
    sib_data_structure* p = (sib_data_structure*) p_param;
    switch (op->header->tr_type)
    {
    case M3_INSERT:
        g_mutex_lock(op->op_lock);
        //g_mutex_lock(p->store_lock);
        qDebug("Beginning to insert for transaction %d\n", op->header->tr_id);
        rdf_writer(op, p);
        qDebug("Done inserting for transaction %d\n", op->header->tr_id);
        //g_mutex_unlock(p->store_lock);
        op->op_complete = TRUE;
        // qDebug("Now signaling transaction %d for finished operation\n", op->header->tr_id);
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;
    case M3_REMOVE:
        g_mutex_lock(op->op_lock);
        //g_mutex_lock(p->store_lock);
        qDebug("Beginning to remove for transaction %d\n", op->header->tr_id);
        rdf_retractor(op, p);
        qDebug("Done removing for transaction %d\n", op->header->tr_id);
        //g_mutex_unlock(p->store_lock);
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;
    case M3_UPDATE:
        g_mutex_lock(op->op_lock);
        //g_mutex_lock(p->store_lock);
        qDebug("Beginning to update for transaction %d\n", op->header->tr_id);
        rdf_retractor(op, p);
        rdf_writer(op, p);
        qDebug("Done updating for transaction %d\n", op->header->tr_id);
        //g_mutex_unlock(p->store_lock);
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;

        /*AD-ARCES*/
    case M3_PROTECTION_FAULT:
        /* SIB PROTECTION FAULT CASE */
        qWarning("----> SIB PROTECTION FAULT CASE\n");
        g_mutex_lock(op->op_lock);
        /*AD-ARCES*/
        op->rsp->status = ss_SIBProtectionFault;// ss_SIBFailAccessDenied;// ss_InvalidParameter;// ss_OperationFailed;//
        /*AD-ARCES*/
        op->op_complete = TRUE; //FALSE; //it is just for lock purpose
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;


    default:
        /* ERROR CASE */
        g_mutex_lock(op->op_lock);
        op->rsp->status = ss_InvalidParameter;
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;
    }
}


void do_query_subscribe(gpointer op_param, gpointer p_param)
{
    scheduler_item* op = (scheduler_item*) op_param;
    sib_data_structure* p = (sib_data_structure*) p_param;


    switch (op->header->tr_type)
    {
    /* Query and subscribe handled similarly at this level (for now) */
    case M3_SUBSCRIBE:
        if (op->rsp->first_subscribe)
        {
            //For the first subscribe a complete query is needed
            g_mutex_lock(op->op_lock);
            //Generating sub TS

            if (op->rsp->sparql_subscribe==TRUE)
            {
                // Necessary to run firstly as a template to create a triple context for
                // the first sparql result (probably the faster solution)
                //////////////////////////////////////
                op->req->type=QueryTypeTemplate;
                rdf_reader(op,p);
                op->req->type=QueryTypeSPARQLSelect;
                //////////////////////////////////////
            }
            else
            {
                rdf_reader(op,p);
            }

            op->rsp->first_subscribe=FALSE;
            threadsub_wake_and_reload_op(op,p);
            g_mutex_unlock(op->op_lock);
        }
        else
        {	//Common query
            g_mutex_lock(op->op_lock);
            rdf_subscribe(op,p);
            g_mutex_unlock(op->op_lock);
        }
        break;
    case M3_QUERY_SPQL_UPD:
        g_mutex_lock(op->op_lock);
        rdf_reader(op,p);
        /////////////////////////////
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        /////////////////////////////
        g_mutex_unlock(op->op_lock);
        //qDebug("processing query");
        break;
    default:
        /* ERROR CASE */
        g_mutex_lock(op->op_lock);
        op->rsp->status = ss_InvalidParameter;
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        g_mutex_unlock(op->op_lock);
        break;
    }
}

void set_sub_to_pending(gpointer sub_id, gpointer sub_data, gpointer unused)
{
    subscription_state* s = (subscription_state*)sub_data;

    if (s->status != M3_SUB_STOPPED)
        s->status = M3_SUB_PENDING;

    qDebug("Set subscription %s to pending\n", s->sub_id); /* SUB_DEBUG */

}

void set_sub_to_ongoing(gpointer sub_id, gpointer sub_data, gpointer unused)
{
    subscription_state* s = (subscription_state*)sub_data;
    if (s->status != M3_SUB_STOPPED)
        s->status = M3_SUB_ONGOING;
    qDebug("Set subscription %s to ongoing\n", s->sub_id); /* SUB_DEBUG */
}

ssStatus_t rdf_reader(scheduler_item* op, sib_data_structure* p)
{
    qDebug("Querying in transaction %d\n", op->header->tr_id);



    gchar*	XML_no_pre;
    gchar*	XML_no_pre_nospace;
    gchar*	result_sparql_query;
    size_t 	length;

    //Rdf Redland Initialize..
    librdf_statement* 	statement ;
    librdf_statement*	partial_statement;
    librdf_stream* stream;
    librdf_stream* stream2;

    librdf_node* sub_node ;
    librdf_node* pred_node;
    librdf_node* obj_node ;

    /////////////////////////


    qDebug("Query Executing (or first subscribe):\n");
    switch (op->req->type)
    {
    case QueryTypeTemplate:
    {
        ssTriple_t* tq;
        GSList* query_list = op->req->template_query;
        ssTriple_t *triple;

        //Clean old Results
        //ssFreeTripleList(op->rsp->results);
        op->rsp->results = NULL;

        //qDebug("Doing template query");
        /* Changed if anything goes wrong */
        op->rsp->status = ss_StatusOK;

        /* Iterate over query here */
        while (query_list != NULL) {


            tq = (ssTriple_t*) query_list->data;
            if (!tq->subject || !tq->predicate || !tq->object)
            {
                op->rsp->status = ss_OperationFailed;
                op->rsp->results = NULL;
                break;
            }


            //Reasoning, prefix transform
            //tq=trasform_statement_contract(tq);
            tq=trasform_statement_expand(tq);
            //////////////////////////////////


            //qDebug("RDF reader: now querying for transaction %d\n", op->header->tr_id);
            //qDebug("RDF reader: triple is \n%s\n%s\n%s\n", tq->subject, tq->predicate, tq->object);


            // Redland query with wildcard support
            partial_statement=librdf_new_statement(p->RDF_world);

            if ( (strcmp((char*) tq->subject,wildcard1) ==0) || ( (strcmp((char*) tq->subject,wildcard2) ==0) ))
            {	librdf_statement_set_subject(partial_statement, NULL);
            }
            else
            {	librdf_statement_set_subject(partial_statement, librdf_new_node_from_uri_string(p->RDF_world,  tq->subject));
            }

            if ( (strcmp((char*) tq->predicate,wildcard1) ==0) || ( (strcmp((char*) tq->predicate,wildcard2) ==0) ))
            {	librdf_statement_set_predicate(partial_statement, NULL);
            }
            else
            {	librdf_statement_set_predicate(partial_statement, librdf_new_node_from_uri_string(p->RDF_world,tq->predicate));
            }

            if ( (strcmp((char*) tq->object,wildcard1) ==0) || ( (strcmp((char*) tq->object,wildcard2) ==0) ))
            {	librdf_statement_set_object(partial_statement, NULL);
            }
            else
            {
                if (! tq->objType)
                {	librdf_statement_set_object(partial_statement,
                        librdf_new_node_from_uri_string(p->RDF_world, tq->object));
                }
                else
                {	librdf_statement_set_object(partial_statement,
                        librdf_new_node_from_literal(p->RDF_world, tq->object, NULL, 0));
                }
            }

            if (p->virtuoso)
                stream=librdf_model_find_statements_in_context(p->RDF_model, partial_statement, p->virtuoso_context);
            else
                stream=librdf_model_find_statements(p->RDF_model, partial_statement);

            librdf_free_statement(partial_statement);


            //fprintf(stderr, "Query print all\n");
            while(!librdf_stream_end(stream)) {
                statement=librdf_stream_get_object(stream);

                //CHARGE FIRST TIME LOCAL TRIPLES IN LOCAL CONTEXT STORE///////////////////////////////////
                if (op->rsp->first_subscribe)
                {
                    // NECESSARY FOR LITERAL^^STRING FROM VIRTUOSO
                    if (p->virtuoso)
                    {
                        librdf_statement * forced_statement = librdf_new_statement_from_statement(statement);
                        if (librdf_node_is_literal(librdf_statement_get_object(statement)))
                        {
                            librdf_free_node(librdf_statement_get_object(forced_statement));
                            librdf_statement_set_object(forced_statement,
                                librdf_new_node_from_literal(p->RDF_world,
                                    librdf_node_get_literal_value(librdf_statement_get_object(statement)), NULL, 0));
                        }
                        librdf_model_add_statement(op->rsp->submodel, forced_statement);
                        librdf_free_statement(forced_statement);
                    }
                    else
                    {
                        librdf_model_add_statement(op->rsp->submodel, statement);
                    }
                }
                ///////////////////////////////////////////////////////////////////////////////////////////

                if (op->rsp->sparql_subscribe != TRUE)
                {
                        sub_node  = librdf_statement_get_subject(statement);
                        pred_node = librdf_statement_get_predicate(statement);
                        obj_node  = librdf_statement_get_object(statement);

                        if (( librdf_node_is_blank(sub_node))||( librdf_node_is_blank(obj_node)))
                        {
                            librdf_stream_next(stream);
                            continue;
                        }

                        triple  =(ssTriple_t *) g_new0(ssTriple_t,1);
                        g_return_val_if_fail(triple, ss_NotEnoughResources);


                        triple->subject=	(ssElement_t) g_strdup((gchar*) librdf_uri_as_string( librdf_node_get_uri(sub_node)));
                        triple->predicate =	(ssElement_t) g_strdup((gchar*) librdf_uri_as_string( librdf_node_get_uri(pred_node)));


                        if (librdf_node_is_literal(obj_node))
                        {
                            triple->object =  (ssElement_t) g_strdup((gchar*) librdf_node_get_literal_value(obj_node));
                            triple->objType = ssElement_TYPE_LIT;
                        }
                        else
                        {
                            triple->object =  (ssElement_t) g_strdup((gchar*) librdf_uri_as_string( librdf_node_get_uri(obj_node)));
                            triple->objType = ssElement_TYPE_URI;
                        }

                        op->rsp->results = g_slist_prepend(op->rsp->results, triple);

                }
                librdf_stream_next(stream);
            }

            librdf_free_stream(stream);
            //
            query_list = query_list->next;
        }
        break;
    }

    // WILBOUR QUERY EMULATOR      TODO?
    /*
        case QueryTypeWQLValues:
          {	      }
        case QueryTypeWQLNodeTypes:
          {	      }
        case QueryTypeWQLRelated:
          {	      }
        case QueryTypeWQLIsType:
          {	      }
        case QueryTypeWQLIsSubType:
          {	      }
     /*  EM WQL TODO */


    case QueryTypeSPARQLSelect:
    {
        op->rsp->status = ss_StatusOK;
        qDebug("RECEIVED SPARQL (query or sub ) IN\n");
        qDebug("q: %s\n",op->req->query_str);


        op->rsp->results_str= (gchar*) load_sparql_on_dedicate_sub_ts((unsigned char*)op->req->query_str, p->RDF_world, p->RDF_storage, p->RDF_model);
        if (g_strcmp0(op->rsp->results_str,"")==0)
        {
            op->rsp->status = ss_GeneralError;
        }
        break;
    }


    //End case query type
    }

    return op->rsp->status;
}

ssStatus_t rdf_writer(scheduler_item* op, sib_data_structure* param)
{

    switch (op->req->encoding)
    {
    case EncodingM3XML:
        qDebug("Writing RDF/M3 in transaction %d\n", op->header->tr_id);
        GSList* i ;
        ssTriple_t* t;

        //Rdf Redland Initialize..
        librdf_statement* statement ;
        ///////////////////////////


        for (i = op->req->insert_graph;
                i != NULL;
                i = i->next)
        {
            t = (ssTriple_t*)i->data;

            //Reasoning, prefix transform
            //t=trasform_statement_contract(t);
            t=trasform_statement_expand(t);
            //////////////////////////////////////////

            //Insert statement
            statement=librdf_new_statement(param->RDF_world);

            librdf_statement_set_subject(statement,
                    librdf_new_node_from_uri_string(param->RDF_world,  t->subject));

            librdf_statement_set_predicate(statement,
                    librdf_new_node_from_uri_string(param->RDF_world, t->predicate));

            if (! t->objType) {
                //URI: 	t->objType==0
                librdf_statement_set_object(statement,
                        librdf_new_node_from_uri_string(param->RDF_world,  t->object));
            }
            else
            {
                //Literal:  	t->objType==1
                librdf_statement_set_object(statement,
                        librdf_new_node_from_literal(param->RDF_world, t->object, NULL, 0));

            }

            if ( param->virtuoso)
                librdf_model_context_add_statement(param->RDF_model, param->virtuoso_context, statement);
            else
                librdf_model_add_statement(param->RDF_model, statement);


            //FOR DIFFERENTIAL SUBSCRIBE
            //g_mutex_lock(param->temp_ins_rem_operations_lock);
            //librdf_model_add_statement(param->RDF_model_insert, statement);

            //qDebug("Inserting\n");
            //librdf_statement_print(statement , stdout);
            //qDebug("\n");

            statement_in_tl(&(param->RDF_list_insert), statement);

            ///////////////////////////////////////Reasoning!
            qDebug("WARNING: REASONING\n");
            reasoning(param,t, param->enable_rdf_pp);
            ////////////////////////////////////////////////
            param->subs_sheduler_inserted_triple=TRUE;
            //g_mutex_unlock(param->temp_ins_rem_operations_lock);

            librdf_free_statement(statement);

        }

        op->rsp->status = ss_StatusOK;
        break;

    case EncodingRDFXML:
    {

        librdf_parser* parser;
        librdf_uri* uri;
        librdf_stream* stream;
        librdf_statement* statement;

        //gboolean * blanks_founded;
        int blanks_founded;

        uri=librdf_new_uri(param->RDF_world, (const unsigned char*) "http://example.librdf.org/");
        parser=librdf_new_parser(param->RDF_world, "rdfxml", NULL, NULL);

        stream=librdf_parser_parse_string_as_stream (parser, (const unsigned char*) op->req->insert_str,uri);

        if (!stream)
        {
            fprintf(stderr, "Failed to parse RDF into model\n");
            op->rsp->status = ss_OperationFailed;
            // No bnodes to uri mapping from RDF/XML content
            // op->rsp->bnodes_str = g_strdup("<urilist></urilist>");
        }
        else
        {
            //FOR DIFFERENTIAL SUBSCRIBE
            //g_mutex_lock(param->temp_ins_rem_operations_lock);
            param->subs_sheduler_inserted_triple=TRUE;
            /////////////////////////////

            while(!librdf_stream_end(stream))
            {
                statement= librdf_stream_get_object(stream);

                blanks_founded=FALSE;

                ssTriple_t *ttemp = (ssTriple_t *)g_new0(ssTriple_t,1);
                g_return_val_if_fail(ttemp, ss_NotEnoughResources);

                if ( ! librdf_node_is_blank(librdf_statement_get_subject(statement)))
                {
                    ttemp->subject=	(ssElement_t) g_strdup((gchar*) librdf_uri_as_string(librdf_node_get_uri(librdf_statement_get_subject(statement))));
                }
                else //Bnode
                {
                    blanks_founded=TRUE;
                }

                ttemp->predicate = (ssElement_t) g_strdup((gchar*) librdf_uri_as_string(librdf_node_get_uri(librdf_statement_get_predicate(statement))));

                if (librdf_node_is_literal(librdf_statement_get_object(statement)))
                {

                    ttemp->object =  (ssElement_t) g_strdup((gchar*) librdf_node_get_literal_value(librdf_statement_get_object(statement)));
                    ttemp->objType = ssElement_TYPE_LIT;
                }
                else if ( ! librdf_node_is_blank(librdf_statement_get_object(statement)))
                {

                    ttemp->object =  (ssElement_t) g_strdup((gchar*) librdf_uri_as_string( librdf_node_get_uri(librdf_statement_get_object(statement))));
                    ttemp->objType = ssElement_TYPE_URI;
                }
                else  //Bnode
                {
                    blanks_founded=TRUE;
                }

                /////////////////////Best way to solve blank nodes issues is not have them at all
                if (! blanks_founded)
                {
                    if ( param->virtuoso)
                        librdf_model_context_add_statement(param->RDF_model, param->virtuoso_context, statement);
                    else
                        librdf_model_add_statement(param->RDF_model, statement);

                    statement_in_tl(&(param->RDF_list_insert), statement);
                    reasoning(param,ttemp, param->enable_rdf_pp);
                }
                //////////////////////////////////////////////////////////////////////////////////

                ssFreeTriple(ttemp);

                librdf_stream_next(stream);
            }
            //g_mutex_unlock(param->temp_ins_rem_operations_lock);

            //fprintf(stderr, "RDF inserted ..\n");
            op->rsp->status = ss_StatusOK;
        }


        librdf_free_stream(stream);
        librdf_free_parser(parser);
        librdf_free_uri(uri);

        break;

    }
    default: /* ERROR CASE */
        op->rsp->status = ss_InvalidParameter;
        //op->rsp->bnodes_str = g_strdup("<urilist></urilist>");
        break;
    }
    return op->rsp->status;
}


ssStatus_t rdf_retractor(scheduler_item* op, sib_data_structure* param)
{
    switch (op->req->encoding)
    {
    case EncodingM3XML:
    {
        qDebug("Removing in transaction %d\n", op->header->tr_id);

        GSList *i;
        ssTriple_t* tq;

        //Rdf Redland Initialize..
        librdf_statement* 	statement ;
        librdf_statement*	partial_statement;
        librdf_stream* 	stream;
        librdf_statement* 	remove_statement_for_subs;
        ///////////////////

        //gboolean* al_least_one_wildcard=0;
        int al_least_one_wildcard=0;
        //gboolean* all_wildcard=0;
        int all_wildcard=0;

        for (i = op->req->remove_graph; i != NULL; i = i->next)
        {
            tq = (ssTriple_t*)i->data;

            //////Reasoning, prefix transform
            //tq=trasform_statement_contract(tq);
            tq=trasform_statement_expand(tq);
            //////////////////////////////////


            al_least_one_wildcard=0;
            all_wildcard=1;

            partial_statement=librdf_new_statement(param->RDF_world);

            if ( (strcmp((char*) tq->subject,wildcard1) ==0) || ( (strcmp((char*) tq->subject,wildcard2) ==0) ))
            {	librdf_statement_set_subject(partial_statement, NULL);
                //fprintf (stderr,"wildcard subj enable ");
                al_least_one_wildcard=1;
            }
            else
            {	librdf_statement_set_subject(partial_statement, librdf_new_node_from_uri_string(param->RDF_world,  tq->subject));
                all_wildcard=0;
                //fprintf (stderr,"no wildcard subj enable");
            }

            if ( (strcmp((char*) tq->predicate,wildcard1) ==0) || ( (strcmp((char*) tq->predicate,wildcard2) ==0) ))
            {	librdf_statement_set_predicate(partial_statement, NULL);
                //fprintf (stderr,va"wildcard pred enable ");
                al_least_one_wildcard=1;
            }
            else
            {	librdf_statement_set_predicate(partial_statement, librdf_new_node_from_uri_string(param->RDF_world,  tq->predicate));
            all_wildcard=0;
            }

            if ( (strcmp((char*) tq->object,wildcard1) ==0) || ( (strcmp((char*) tq->object,wildcard2) ==0) ))
            {	librdf_statement_set_object(partial_statement, NULL);
                //fprintf (stderr,"wildcard obj enable ");
                al_least_one_wildcard=1;
            }
            else
            {
                all_wildcard=0;
                if (! tq->objType)
                {
                    librdf_statement_set_object(partial_statement,
                            librdf_new_node_from_uri_string(param->RDF_world,  tq->object));
                }
                else
                {
                    librdf_statement_set_object(partial_statement,
                            librdf_new_node_from_literal(param->RDF_world,tq->object, NULL, 0));
                }
            }

            param->subs_sheduler_removed_triple=TRUE;

            if (al_least_one_wildcard)
            {
                if (all_wildcard)
                {
                    //If the TS is big I can't copy all the content in ram.. so I need this solution.
                    qDebug("ALL WC! RESET DB\n");
                    reset_storage_model( param);
                }
                else
                {
                    //One or more WC
                    if (param->virtuoso)
                        stream=librdf_model_find_statements_in_context(param->RDF_model, partial_statement,  param->virtuoso_context);
                    else
                        stream=librdf_model_find_statements(param->RDF_model, partial_statement);

                    while(!librdf_stream_end(stream))
                    {
                        statement=librdf_stream_get_object(stream);

                        //qDebug("Removing\n");
                        //librdf_statement_print(statement , stdout);
                        //qDebug("\n");

                        ////////////////
                        statement_in_tl(&(param->RDF_list_remove), statement);
                        ////////////////

                        if (param->virtuoso)
                            librdf_model_context_remove_statement(param->RDF_model, param->virtuoso_context, statement);
                        else
                            librdf_model_remove_statement(param->RDF_model, statement);

                        librdf_stream_next(stream);

                    }
                    librdf_free_stream(stream);
                }
            }
            else
            {
                //No WC
                ////////////////
                statement_in_tl(&(param->RDF_list_remove), partial_statement);
                ///////////////

                //qDebug("Removing\n");
                //librdf_statement_print(partial_statement , stdout);
                //qDebug("\n");

                if (param->virtuoso)
                    librdf_model_context_remove_statement(param->RDF_model, param->virtuoso_context, partial_statement);
                else
                    librdf_model_remove_statement(param->RDF_model,partial_statement);

            }

            librdf_free_statement(partial_statement);

        }

        op->rsp->status = ss_StatusOK;
        return op->rsp->status;

    }
    case EncodingRDFXML:
    {
        librdf_parser* parser;
        librdf_uri* uri;
        librdf_stream* stream;
        librdf_statement* statement ;


        uri=librdf_new_uri(param->RDF_world, (const unsigned char*) "http://example.librdf.org/");
        parser=librdf_new_parser(param->RDF_world, "rdfxml", NULL, NULL);

        stream=librdf_parser_parse_string_as_stream (parser, (const unsigned char*) op->req->remove_str,uri);

        if (!stream)
        {
            fprintf(stderr, "Failed to parse RDF into model\n");
            op->rsp->status = ss_OperationFailed;
            // No bnodes to uri mapping from RDF/XML content
            // op->rsp->bnodes_str = g_strdup("<urilist></urilist>");
        }
        else
        {

            //g_mutex_lock(param->temp_ins_rem_operations_lock);
            param->subs_sheduler_inserted_triple=TRUE;

            while(!librdf_stream_end(stream))
            {
                statement= librdf_stream_get_object(stream);

                if (param->virtuoso)
                    librdf_model_context_remove_statement(param->RDF_model, param->virtuoso_context, statement);
                else
                    librdf_model_remove_statement(param->RDF_model, statement);

                statement_in_tl(&(param->RDF_list_remove), statement);
                //librdf_model_add_statement(param->RDF_model_remove, statement);
                librdf_stream_next(stream);
            }
            //g_mutex_unlock(param->temp_ins_rem_operations_lock);


            fprintf(stderr, "RDF removed ..\n");
            op->rsp->status = ss_StatusOK;
        }


        librdf_free_stream(stream);
        librdf_free_parser(parser);
        librdf_free_uri(uri);

        return op->rsp->status;
    }
    }
}

void rdf_subscribe(scheduler_item* op, sib_data_structure *param)
{
    gboolean sub_ts_could_change=FALSE;
    gboolean sub_ts_will_change=FALSE;
    gboolean try_wake_thread=FALSE;
    gboolean sub_processing_on_thread_localvar;

    sib_data_structure* p = param;

    if (threadsub_wake_if_unsub(op,p))
    {
        //qDebug("UNSUBSCRIBE ARRIVED, THREADS WAKED! \n");
        return;
    }

    //CHECK IF A THREAD IS AVAILABLE

    g_mutex_lock(p->sub_threads_lock);
    int sub_threads_active_local;
    int sub_threads_max_local;
    sub_threads_active_local=p->sub_threads_active;
    sub_threads_max_local=p->sub_threads_max;
    g_mutex_unlock(p->sub_threads_lock);


    if (sub_threads_active_local  < sub_threads_max_local)
    {
        //qDebug("p->sub_threads_active %d on %d\n",sub_threads_active_local ,sub_threads_max_local );

        g_mutex_lock(op->rsp->sub_processing_on_thread_lock);
        op->rsp->sub_processing_on_thread=TRUE;
        sub_processing_on_thread_localvar=TRUE;
        g_mutex_unlock(op->rsp->sub_processing_on_thread_lock);
    }
    else
    {
        gboolean thread_busy_local;
        g_mutex_lock(op->thread_loop_lock);
        thread_busy_local=op->thread_busy;
        g_mutex_unlock(op->thread_loop_lock);

        while (thread_busy_local == TRUE)
        {
            //Not so elegant for now, but it works fine
            usleep(250);

            g_mutex_lock(op->thread_loop_lock);
            thread_busy_local=op->thread_busy;
            g_mutex_unlock(op->thread_loop_lock);
        }

        g_mutex_lock(op->rsp->sub_processing_on_thread_lock);
        op->rsp->sub_processing_on_thread=FALSE;
        sub_processing_on_thread_localvar=FALSE;
        g_mutex_unlock(op->rsp->sub_processing_on_thread_lock);

    }

    sub_ts_could_change = load_seq_i_r_to_async_ts(op, p);
    if (sub_ts_could_change)
    {
        if (! sub_processing_on_thread_localvar)
        {
            if (op->rsp->sparql_subscribe)	//SPARQL SUBSCRIBE
            {	try_wake_thread=load_sparql_booster_subscriptions(op,p);
            }
            else							//RDF SUBSCRIBE
            {   try_wake_thread=load_rdf_subscriptions(op,p);
            }

            threadsub_try_not_wake_reload_op(op, p);
        }
        else
        {	threadsub_wake_and_reload_op(op, p);
        }
    }
    else // the ins/rm triples will not change sub-TS (in both cases RDF / SPARQL), check for unsubs
        threadsub_try_not_wake_reload_op(op, p);

}

gboolean load_rdf_subscriptions(scheduler_item* op, sib_data_structure* sib_data)
{
    gchar *  added_str      =NULL;
    gchar *  removed_str    =NULL;
    gboolean value = FALSE;

    sib_data_structure* p = sib_data;

    while (check_add_rm_on_substore_from_i_r(op->req, op->rsp ))
    {
        //qDebug("Loading rdf sub_seq\n");

        if ((op->rsp->added == NULL) && (op->rsp->removed == NULL))
        {
            value= value | FALSE;
            continue;
        }

        value= value | TRUE;

        add_triplelist_in_submodel    (p, op->rsp, op->rsp->added);
        remove_triplelist_in_submodel (p, op->rsp, op->rsp->removed);

        added_str=m3_gen_triple_string(op->rsp->added);
        removed_str=m3_gen_triple_string(op->rsp->removed);

        if( ++(op->rsp->ind_seqnum) == SSAP_IND_WRAP_NUM )
            op->rsp->ind_seqnum=1;

/*
        sib_op_parameter * sib_op_param = op->rsp->param;

        whiteboard_util_send_signal(SIB_DBUS_OBJECT,
                SIB_DBUS_KP_INTERFACE,
                SIB_DBUS_KP_SIGNAL_SUBSCRIPTION_IND,
                sib_op_param->conn,
                DBUS_TYPE_STRING, &(op->header->space_id),
                DBUS_TYPE_STRING, &(op->header->kp_id),
                DBUS_TYPE_INT32, &(op->header->tr_id),
                DBUS_TYPE_INT32, &(op->rsp->ind_seqnum),
                DBUS_TYPE_STRING, &(op->rsp->sub_id),
                DBUS_TYPE_STRING, &(added_str),
                DBUS_TYPE_STRING, &(removed_str),
                WHITEBOARD_UTIL_LIST_END);
*/
        qDebug(" ********** SUBSCRIPTION IND!\n");
        handle_subscription_ind(op->header->space_id, op->header->kp_id, op->header->tr_id,
                                op->rsp->ind_seqnum, op->rsp->sub_id, added_str, removed_str);

        g_free(added_str);
        g_free(removed_str);

        m3_free_triple_list_simple(&(op->rsp->added));
        m3_free_triple_list_simple(&(op->rsp->removed));

    }

    return value;

}

gboolean threadsub_wake_if_unsub(scheduler_item* op, sib_data_structure* p)
{
    //CHECK IF UNSUBSCRIBED!
    subscription_state* s;
    g_mutex_lock(p->subscriptions_lock);
    s = (subscription_state*) g_hash_table_lookup(p->subs, op->rsp->sub_id);
    if ((s->status == M3_SUB_STOPPED) || (s==NULL))
    {
        g_mutex_unlock(p->subscriptions_lock);

        /////////////////////////////
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        /////////////////////////////

        return TRUE;
    }
    else
    {
        g_mutex_unlock(p->subscriptions_lock);
        return FALSE;
    }
}

void threadsub_wake_and_reload_op(scheduler_item* op, sib_data_structure* p)
{
    //qDebug("Try Waking the Thread \n");

    //Trigger for unsub
    subscription_state* s;
    g_mutex_lock(p->subscriptions_lock);
    s = (subscription_state*) g_hash_table_lookup(p->subs, op->rsp->sub_id);
    if (NULL != s && s->status != M3_SUB_STOPPED)
    {
        s->status = M3_SUB_ONGOING;
        g_mutex_unlock(p->subscriptions_lock);
        //qDebug("QUERY FOR SUBS: Set subscription %s to ongoing\n", s->sub_id); /* SUB_DEBUG */

        g_async_queue_push(p->query_queue, op);
    }
    else if ((s->status == M3_SUB_STOPPED) || (NULL == s))
    {
        g_mutex_unlock(p->subscriptions_lock);
    }
    //If the thread is busy no signal is required
    g_mutex_lock(op->thread_loop_lock);
    if (op->thread_busy==FALSE)
    {
        /////////////////////////////
        op->op_complete = TRUE;
        g_cond_signal(op->op_cond);
        /////////////////////////////
    }
    g_mutex_unlock(op->thread_loop_lock);
}

void threadsub_try_not_wake_reload_op(scheduler_item* op, sib_data_structure* p)
{
    //Not necessary to wake the thread, just in case of unsubs
    //If no unsub happens the op will be autorecharged

    ///////////////////////////////////////////////////////////
    //Trigger for unsub

    subscription_state* s;
    g_mutex_lock(p->subscriptions_lock);
    s = (subscription_state*) g_hash_table_lookup(p->subs, op->rsp->sub_id);
    if ((NULL != s && s->status == M3_SUB_STOPPED) || (NULL == s))
    {
        g_mutex_unlock(p->subscriptions_lock);

        ////////////////////////////////////////
        //qDebug("cleaning submodels\n");
        //If the thread is busy no signal is required
        g_mutex_lock(op->thread_loop_lock);
        if (op->thread_busy==FALSE)
        {
            //qDebug("seq to thread:UNSUB\n");

            /////////////////////////////
            op->op_complete = TRUE;
            g_cond_signal(op->op_cond);
            /////////////////////////////

        }
        g_mutex_unlock(op->thread_loop_lock);

    }
    else //SUB still active, the persistent query will be auto-recharged
    {
        g_mutex_unlock(p->subscriptions_lock);
        g_async_queue_push(p->query_queue, op);
    }
    ////////////////////////////////////////////////////////


}

gboolean  load_seq_i_r_to_async_ts(scheduler_item* op, sib_data_structure* p)
{

    gboolean result =FALSE;

    gboolean  same_s;
    gboolean  same_p;
    gboolean  same_o;

    GSList * transaction_added_triples = NULL;
    GSList * transaction_removed_triples = NULL;

    added_removed_lists_struct * add_rm_transaction =g_new0 (added_removed_lists_struct,1);

    //Check if something has been added or removed..
    if (p->subs_sheduler_inserted_triple || p->subs_sheduler_removed_triple)
    {

        ssTriple_t_sparql* tq;
        GSList* query_list = op->req->template_query;

        /* Iterate over query here */
        while (query_list != NULL)
        {
            tq = (ssTriple_t_sparql*) query_list->data;
            if (!tq->subject || !tq->predicate || !tq->object)
            {
                op->rsp->status = ss_OperationFailed;
                op->rsp->results = NULL;
                break;
            }

            //////////////////FILTER OUT IDENTICAL QUERY PATTERNS//////////////////////
            if (op->rsp->sparql_subscribe)
            {
                if (tq->replicated)
                {
                    query_list=query_list->next;
                    continue;
                }
            }

            //////////////////END FILTER IDENTICAL QUERY PATTERNS//////////////////////
            if ((p->subs_sheduler_removed_triple ))
            {
                GSList * pointer_removed = p->RDF_list_remove;
                while (pointer_removed)
                {
                    ssTriple_t * removed = (ssTriple_t *) pointer_removed->data;

                    same_s=FALSE;
                    same_p=FALSE;
                    same_o=FALSE;

                    if (
                            (strcmp((char*) removed->subject,(char*) tq->subject) ==0) ||
                            (strcmp((char*) tq->subject,wildcard1) ==0) ||
                            (strcmp((char*) tq->subject,wildcard2) ==0)
                    )
                    {
                        same_s=TRUE;
                    }
                    else
                    {
                        pointer_removed=pointer_removed->next;
                        continue;
                    }

                    if (
                            (strcmp((char*) removed->predicate,(char*) tq->predicate) ==0) ||
                            (strcmp((char*) tq->predicate,wildcard1) ==0) ||
                            (strcmp((char*) tq->predicate,wildcard2) ==0)
                    )
                    {
                        same_p=TRUE;
                    }
                    else
                    {
                        pointer_removed=pointer_removed->next;
                        continue;
                    }

                    if (
                            ((strcmp((char*) removed->object,(char*) tq->object) == 0) && ((removed->objType)==(tq->objType))) ||
                            (strcmp((char*) tq->object,wildcard1) ==0) ||
                            (strcmp((char*) tq->object,wildcard2) ==0)
                    )
                    {
                        same_o=TRUE;
                    }
                    else
                    {
                        pointer_removed=pointer_removed->next;
                        continue;
                    }

                    if (same_s && same_p && same_o)
                        //No wildcard on statement removed
                        {
                        ssTriple_t * copied_triple = ssTripleHardCopy(removed);

                        //qDebug("Removing triple from triplelist potentially removed \n");
                        //qDebug("%s %s %s %d\n",copied_triple->subject , copied_triple->predicate, copied_triple->object , copied_triple-> objType);

                        //g_async_queue_lock(op->rsp->insert_remove_queue_async);
                        transaction_removed_triples = g_slist_prepend(transaction_removed_triples,copied_triple);
                        //g_async_queue_unlock(op->rsp->insert_remove_queue_async);

                        result=TRUE;
                        }

                    pointer_removed=pointer_removed->next;
                }
            }

            if (p->subs_sheduler_inserted_triple )
            {
                GSList * pointer_added = p->RDF_list_insert;
                while (pointer_added)
                {
                    ssTriple_t * added = (ssTriple_t *) pointer_added->data;

                    same_s=FALSE;
                    same_p=FALSE;
                    same_o=FALSE;

                    if (
                            (strcmp((char*) added->subject,(char*) tq->subject) ==0) ||
                            (strcmp((char*) tq->subject,wildcard1) ==0) ||
                            (strcmp((char*) tq->subject,wildcard2) ==0)
                    )
                    {
                        same_s=TRUE;
                    }
                    else
                    {
                        pointer_added=pointer_added->next;
                        continue;
                    }

                    if (
                            (strcmp((char*) added->predicate,(char*) tq->predicate) ==0) ||
                            (strcmp((char*) tq->predicate,wildcard1) ==0) ||
                            (strcmp((char*) tq->predicate,wildcard2) ==0)
                    )
                    {
                        same_p=TRUE;
                    }
                    else
                    {
                        pointer_added=pointer_added->next;
                        continue;
                    }

                    if (
                            ((strcmp((char*) added->object,(char*) tq->object) == 0) && ((added->objType)==(tq->objType))) ||
                            (strcmp((char*) tq->object,wildcard1) ==0) ||
                            (strcmp((char*) tq->object,wildcard2) ==0)
                    )
                    {
                        same_o=TRUE;
                    }
                    else
                    {
                        pointer_added=pointer_added->next;
                        continue;
                    }

                    if (same_s && same_p && same_o)
                        //No wildcard on statement removed
                        {

                        ssTriple_t * copied_triple = ssTripleHardCopy(added);

                        //qDebug("Adding triple from triplelist potentially added \n");
                        //qDebug("%s %s %s %d \n",copied_triple->subject ,  copied_triple->predicate, copied_triple->object , copied_triple-> objType);

                        //g_async_queue_lock(op->rsp->insert_remove_queue_async);
                        transaction_added_triples = g_slist_prepend(transaction_added_triples,copied_triple);
                        //g_async_queue_unlock(op->rsp->insert_remove_queue_async);

                        result=TRUE;
                        }

                    pointer_added=pointer_added->next;
                }

            }

            query_list = query_list->next;
        } //End while
    } // End something added or removed

    ///////////////////////////////////////

    if (result)
    {
        //g_async_queue_lock(op->rsp->insert_remove_queue_async);
        add_rm_transaction->removed_list= transaction_removed_triples;
        add_rm_transaction->added_list= transaction_added_triples;
        g_async_queue_push(op->rsp->insert_remove_queue_async, add_rm_transaction);
        //g_async_queue_push_unlocked(op->rsp->insert_remove_queue_async, add_rm_transaction);
        //g_async_queue_unlock(op->rsp->insert_remove_queue_async);
    }
    else
    {
        g_free(add_rm_transaction);
    }

    return (result);

}

gboolean load_sparql_booster_subscriptions(scheduler_item* op, sib_data_structure *p)
{
    gchar *  added_str      =NULL;
    gchar *  removed_str    =NULL;
    gboolean value = FALSE;

    //Time utilities
    struct timeval start, end;
    unsigned long mtime, seconds, useconds, global_t =0;


    while (check_add_rm_on_substore_from_i_r(op->req, op->rsp ))
    {

        //qDebug("SUBSTORAGE NOT UPDATED\n");
        //librdf_model_print(op->rsp->submodel,stdout);


//        if (p->debug_latency)
//        {	//TIME MEASUREMENT UTILITY LINES
//    		gettimeofday(&start, NULL);
//    		/////////////////////////////
//        }
        sparql_booster(op->req,op->rsp,1);
//        if (p->debug_latency)
//        {
//    		///////////////////////////////////////
//    		gettimeofday(&end, NULL);
//    		seconds  = end.tv_sec  - start.tv_sec;
//    		useconds = end.tv_usec - start.tv_usec;
//    		fprintf(stderr,"Booster Time Added [us,sub_id] : %ld , %s \n",  seconds * 1000000 + useconds, op->rsp->sub_id);
//    		///////////////////////////////////////
//        }


//        if (p->debug_latency)
//        {	//TIME MEASUREMENT UTILITY LINES
//    		gettimeofday(&start, NULL);
//    		/////////////////////////////
//        }
        remove_triplelist_in_submodel (p, op->rsp, op->rsp->removed);
        add_triplelist_in_submodel    (p, op->rsp, op->rsp->added);
//        if (p->debug_latency)
//        {
//    		///////////////////////////////////////
//    		gettimeofday(&end, NULL);
//    		seconds  = end.tv_sec  - start.tv_sec;
//    		useconds = end.tv_usec - start.tv_usec;
//    		fprintf(stderr,"Updating sub ts[us,sub_id] : %ld , %s \n",  seconds * 1000000 + useconds, op->rsp->sub_id);
//    		///////////////////////////////////////
//        }


//        if (p->debug_latency)
//        {	//TIME MEASUREMENT UTILITY LINES
//    		gettimeofday(&start, NULL);
//    		/////////////////////////////
//        }
        sparql_booster(op->req,op->rsp,0);
//        if (p->debug_latency)
//        {
//    		///////////////////////////////////////
//    		gettimeofday(&end, NULL);
//    		seconds  = end.tv_sec  - start.tv_sec;
//    		useconds = end.tv_usec - start.tv_usec;
//    		fprintf(stderr,"Booster Time Removed [us,sub_id] : %ld , %s \n",  seconds * 1000000 + useconds, op->rsp->sub_id);
//    		///////////////////////////////////////
//        }



        if ((op->rsp->added_booster == NULL) && (op->rsp->removed_booster ==NULL))
        {
            m3_free_triple_list_simple(&(op->rsp->added));
            m3_free_triple_list_simple(&(op->rsp->removed));

            value= value | FALSE;
            continue;
        }

        value= value | TRUE;

        added_str   = turn_sparql_booster_results_to_str(op->rsp->added_booster  ,  op->rsp->required_sparql_vars);
        removed_str = turn_sparql_booster_results_to_str(op->rsp->removed_booster,  op->rsp->required_sparql_vars);

        if( ++(op->rsp->ind_seqnum) == SSAP_IND_WRAP_NUM )
            op->rsp->ind_seqnum=1;


        sib_op_parameter * sib_op_param = (sib_op_parameter *) op->rsp->param;

        /*
        whiteboard_util_send_signal(SIB_DBUS_OBJECT,
                SIB_DBUS_KP_INTERFACE,
                SIB_DBUS_KP_SIGNAL_SUBSCRIPTION_IND,
                sib_op_param->conn,
                DBUS_TYPE_STRING, &(op->header->space_id),
                DBUS_TYPE_STRING, &(op->header->kp_id),
                DBUS_TYPE_INT32, &(op->header->tr_id),
                DBUS_TYPE_INT32, &(op->rsp->ind_seqnum),
                DBUS_TYPE_STRING, &(op->rsp->sub_id),
                DBUS_TYPE_STRING, &(added_str),
                DBUS_TYPE_STRING, &(removed_str),
                WHITEBOARD_UTIL_LIST_END);
                */
        qDebug(" ********** SUBSCRIPTION IND!\n");

        handle_subscription_ind(op->header->space_id, op->header->kp_id, op->header->tr_id,
                                op->rsp->ind_seqnum, op->rsp->sub_id, added_str, removed_str);

        m3_free_triple_list_simple(&(op->rsp->added));
        m3_free_triple_list_simple(&(op->rsp->removed));

        g_free(added_str);
        g_free(removed_str);

        clean_sparql_booster_results(&(op->rsp->added_booster));
        clean_sparql_booster_results(&(op->rsp->removed_booster));

    }

    return value;

}

gboolean check_add_rm_on_substore_from_i_r(ssap_kp_message *req, ssap_sib_message* rsp )
{

    //qDebug ("check_add_rm_on_substore_from_i_r\n");
    gboolean sub_ts_has_changed = FALSE;
    gboolean queue_ended = FALSE;

    GSList * added_tlist = NULL;
    GSList * pointer_added_tlist = NULL;
    ssTriple_t * triple_added =NULL;

    GSList * removed_tlist = NULL;
    GSList * pointer_removed_tlist = NULL;
    ssTriple_t * triple_removed =NULL;

    librdf_statement * statement_added;
    librdf_statement * statement_removed;

    added_removed_lists_struct * add_rm_transaction = NULL;

    add_rm_transaction= (added_removed_lists_struct *) g_async_queue_try_pop(rsp->insert_remove_queue_async);

    //g_async_queue_lock(rsp->insert_remove_queue_async);
    //add_rm_transaction=g_async_queue_try_pop_unlocked(rsp->insert_remove_queue_async);
    //g_async_queue_unlock(rsp->insert_remove_queue_async);

    if  (!add_rm_transaction )
    {    return FALSE;
    }
    else
    {   queue_ended=TRUE;
    }

    removed_tlist=add_rm_transaction->removed_list;
    added_tlist=add_rm_transaction->added_list;

    pointer_removed_tlist=removed_tlist;
    while (pointer_removed_tlist)
    {

        triple_removed= (ssTriple_t *) pointer_removed_tlist->data;

        ////////CHECKIG FOR IDENTICAL TRIPLES INSERTED AND REMOVED IN THE SAME UPDATE///
        gboolean found_in_added = false;
        pointer_added_tlist=add_rm_transaction->added_list;
        while (pointer_added_tlist)
        {
            triple_added= (ssTriple_t *) pointer_added_tlist->data;
            if  (
                (g_strcmp0((char*) triple_added->subject, (char*) triple_removed->subject)==0)&
                (g_strcmp0((char*) triple_added->predicate, (char*) triple_removed->predicate)==0)&
                (g_strcmp0((char*) triple_added->object, (char*) triple_removed->object)==0)&
                (triple_added->objType==triple_removed->objType)
                )
            {
                found_in_added = true;
                break;
            }
            pointer_added_tlist=pointer_added_tlist->next;
        }
        //////////////////////////////////////////////////////////////////////////////

        if (! found_in_added)
        {
            statement_removed=librdf_new_statement(rsp->subworld);
            librdf_statement_set_subject(statement_removed, librdf_new_node_from_uri_string(rsp->subworld,  triple_removed->subject));
            librdf_statement_set_predicate(statement_removed, librdf_new_node_from_uri_string(rsp->subworld,  triple_removed->predicate));

            if (! triple_removed->objType)
            {
                librdf_statement_set_object(statement_removed,
                        librdf_new_node_from_uri_string(rsp->subworld, triple_removed->object));
            }
            else
            {
                librdf_statement_set_object(statement_removed,
                        librdf_new_node_from_literal(rsp->subworld, triple_removed->object, NULL, 0));
            }


            librdf_stream * stream_tmp = NULL;
            stream_tmp =librdf_model_find_statements(rsp->submodel,statement_removed);

            //If the triple is present, it will be removed for sure.
            //REDLAND BUG; USING STREAM
            if (!librdf_stream_end(stream_tmp))
            {
                // really removed
                //sub_ts_has_changed=TRUE;

                //qDebug("Statement that will be really removed:\n");
                //librdf_statement_print(statement_removed, stdout);
                //qDebug("\n");

                //g_mutex_lock  (op->rsp->add_rm_lock);

                statement_in_tl(&(rsp->removed),statement_removed);
                //g_mutex_unlock  (op->rsp->add_rm_lock);
            }
            librdf_free_statement(statement_removed);
            librdf_free_stream(stream_tmp);

        }

        ssFreeTriple(triple_removed);
        pointer_removed_tlist=pointer_removed_tlist->next;
    }
    g_slist_free(removed_tlist);


    pointer_added_tlist=added_tlist;
    while (pointer_added_tlist)
    {
        triple_added = (ssTriple_t *) pointer_added_tlist->data;

        statement_added=librdf_new_statement(rsp->subworld);

        librdf_statement_set_subject(statement_added, librdf_new_node_from_uri_string(rsp->subworld,  triple_added->subject));
        librdf_statement_set_predicate(statement_added, librdf_new_node_from_uri_string(rsp->subworld,  triple_added->predicate));
        if (! triple_added->objType)
        {
            librdf_statement_set_object(statement_added,
                    librdf_new_node_from_uri_string(rsp->subworld, triple_added->object));
        }
        else
        {
            librdf_statement_set_object(statement_added,
                    librdf_new_node_from_literal(rsp->subworld, triple_added->object, NULL, 0));
        }

        //If the triple is not present, it will be added for sure.
        if (librdf_model_contains_statement(rsp->submodel,statement_added)==0)
        {
            // really removed
            //sub_ts_has_changed=TRUE;

            //qDebug("Statement that will be really added:\n");
            //librdf_statement_print(statement_added, stdout);
            //qDebug("\n");

            //g_mutex_lock  (op->rsp->add_rm_lock);
            statement_in_tl(&(rsp->added),statement_added);
            //g_mutex_unlock  (op->rsp->add_rm_lock);
        }

        librdf_free_statement(statement_added);

        ssFreeTriple (triple_added);
        pointer_added_tlist=pointer_added_tlist->next;
    }
    g_slist_free(added_tlist);


    g_free(add_rm_transaction);
    return queue_ended;

}

void destroy_recycle_world_sub_ts(ssap_sib_message* rsp, sib_data_structure* p)
{

    //Incredibly faster than destroying model-...
    /*
    librdf_stream * stream;
    stream= librdf_model_as_stream(rsp->submodel);
    while(!librdf_stream_end(stream))
    {
        librdf_model_remove_statement(rsp->submodel, librdf_stream_get_object(stream));
        librdf_stream_next(stream);
    }
    librdf_free_stream(stream);
    */

    librdf_free_model(rsp->submodel);
    librdf_free_storage(rsp->substorage);

    // UNRECYCLING STRATEGY
    //librdf_free_world(rsp->subworld);

    // RECYCLING WORLD STRATEGY

    //rsp->submodel=NULL;
    //rsp->substorage=NULL;
    //rsp->subworld=NULL;

    g_mutex_lock(p->sub_threads_lock);
    int sub_threads_max_local;
    sub_threads_max_local=p->sub_threads_max;
    g_mutex_unlock(p->sub_threads_lock);

    if (sub_threads_max_local != 0)
    {
        g_async_queue_lock(p->used_worlds_async_queue);
        g_async_queue_push_unlocked(p->used_worlds_async_queue, rsp->subworld);
        g_async_queue_unlock(p->used_worlds_async_queue);
    }
    // RECYCLING WORLD, MORE ECOLOGIC

}

void generate_recycle_world_sub_ts(ssap_sib_message* rsp, sib_data_structure* p)
{

    g_mutex_lock(p->sub_threads_lock);
    int sub_threads_max_local;
    sub_threads_max_local=p->sub_threads_max;
    g_mutex_unlock(p->sub_threads_lock);

    if (sub_threads_max_local == 0)
    {
        rsp->subworld=p->RDF_world;
    }
    else
    {
        // RECYCLING WORLD, MORE ECOLOGIC
        librdf_world * subworld = NULL;

        g_async_queue_lock(p->used_worlds_async_queue);
        subworld = (librdf_world*) g_async_queue_try_pop_unlocked(p->used_worlds_async_queue);
        g_async_queue_unlock(p->used_worlds_async_queue);

        if (subworld==NULL)
        {
            //qDebug("NON Recycling world...\n");

            rsp->subworld = librdf_new_world();
            librdf_world_open(rsp->subworld );

            //int * librdf_log_func ;
            rasqal_world_set_warning_level(librdf_world_get_rasqal(rsp->subworld), 0);
        }
        else
        {
            //qDebug("Recycling world...\n");
            rsp->subworld =subworld;
        }
        // RECYCLING WORLDs, MORE ECOLOGIC
    }
    //rsp->subworld = librdf_new_world();
    //librdf_world_open(rsp->subworld );

    //////////////////////////////
    librdf_storage * substorage;
    librdf_model * submodel;

    if (!p->subs_persistent)
    {
        if (p->subs_hash)
               substorage=librdf_new_storage(rsp->subworld, "hashes", NULL, "hash-type='memory'");
        else
               substorage=librdf_new_storage(rsp->subworld,  NULL, NULL, NULL);
    }
    else
    {
        gchar** split_result_array;
        gchar** pointer_char_arr;
        gchar * tempname;

        split_result_array= g_strsplit(rsp->sub_id,subs_uri, 0);
        pointer_char_arr=split_result_array;
        pointer_char_arr++;
        tempname=g_strdup_printf("subts_%s",*pointer_char_arr);

        substorage=librdf_new_storage(rsp->subworld, "hashes", tempname,"new='yes',hash-type='bdb',dir='.'");

        g_free (tempname);
        g_strfreev (split_result_array);
    }

    //Initialize  Node with subscription name
    rsp->substorage=substorage;

    submodel=librdf_new_model(rsp->subworld, rsp->substorage, NULL);

    rsp->submodel=submodel;
}

void handle_subscription_ind(gchar *spaceid,
        gchar *nodeid,
        gint msgnum,
        gint seqnum,
        gchar *subscription_id,
        gchar *results_new,
        gchar *results_obsolete)
{

    NodeMsgContent_t* rsp = NULL;
    ssBufDesc_t *bd = NULL;

    rsp = nodemsgcontent_new_subscription_ind((guchar *) spaceid, (guchar *) nodeid, msgnum, seqnum, (guchar *)subscription_id, (guchar *)results_new, (guchar *)results_obsolete);

    if(rsp)
    {
        bd = ssBufDesc_new();


        if( ss_StatusOK != ssBufDesc_CreateSubscriptionIndMessage(bd,
                                                                  (guchar *)parseSSAPmsg_get_nodeid(rsp),
                                                                  (guchar *)parseSSAPmsg_get_spaceid(rsp),
                                                                  parseSSAPmsg_get_msgnumber(rsp),
                                                                  parseSSAPmsg_get_update_sequence (rsp),
                                                                  (guchar *)parseSSAPmsg_get_subscriptionid(rsp),
                                                                  (guchar *)parseSSAPmsg_get_results_added(rsp),
                                                                  (guchar *)parseSSAPmsg_get_results_removed(rsp)))
        {
            ssBufDesc_free(&bd);
        }
    }


 /*
    else
    {

        whiteboard_log_debug("Waiter w/ subscription id: %s not found\n", subscription_id);

        // UNSUBSCRIBE!
        /*
        if( sib_object_send_unsubscribe(listener->object,
                (guchar *)spaceid,
                (guchar *)nodeid,
                1,
                (guchar *)subscription_id) <0)
        {
            whiteboard_log_debug("Could not send unsubscribe request w/ msgid : %d\n",1);
        }
 */

    if(bd)
    {
        subs_ind_t *subs = (subs_ind_t *) malloc(sizeof(subs_ind_t));
        subs->ssap_message = strdup((char*) ssBufDesc_GetMessage(bd));
        subs->node_id = strdup((char *) nodeid);
        subs->subs_id = strdup((char *) subscription_id);

        add_subs_ind(subs);
        //return (char*) ssBufDesc_GetMessage(bd);
    }

    qDebug("SUBSCRIPTION INDICATION!\n");

}

/***** Subscription indication *****/
GList *g_subs_indications = NULL;
GMutex g_subs_ind_lock;

static IndicatorMessanger indMessanger;

void add_subs_ind(subs_ind_t *ind)
{

    indMessanger.sendIndicationEvent(ind);
//    if (g_subs_indications == NULL) {
//        g_subs_indications = g_list_alloc();
//    }

//    g_mutex_lock(&g_subs_ind_lock);
//    g_subs_indications = g_list_append(g_subs_indications, ind);
//    g_mutex_unlock(&g_subs_ind_lock);
}


subs_ind_t *get_subs_ind()
{
    subs_ind_t *ind = NULL;

    g_mutex_lock(&g_subs_ind_lock);
    GList *first = g_list_first(g_subs_indications);

    if (first != NULL) {
        ind = (subs_ind_t *) first->data;
        g_subs_indications = g_list_delete_link(g_subs_indications, first);
    }
    g_mutex_unlock(&g_subs_ind_lock);

    return ind;
}


IndicatorMessanger::IndicatorMessanger(QObject *parent)
{

}

IndicatorMessanger::~IndicatorMessanger()
{

}

void IndicatorMessanger::sendIndicationEvent(subs_ind_t *ind)
{
    emit indicationReady(ind);
}


IndicatorMessanger *getIndcationMessnger()
{
    return &indMessanger;
}
