#include "sibconfig.h"

#include <QDebug>
#include "sibdebug.h"

SibConfig::SibConfig(QString path) : settings(path, QSettings::IniFormat)
{
    qDebug() << "Load config:" << path;

    foreach (QString groupName, settings.childGroups()) {
        if (groupName.startsWith(CUTESIB_CONFIG_AP_PREFIX)) {
            _accessPoints.append(groupName.remove(CUTESIB_CONFIG_AP_PREFIX));
        }
    }

    qDebug() << "Configured access points: " << _accessPoints;
}

SibConfig::~SibConfig()
{

}

QStringList SibConfig::getAccessPoints() const
{
    return _accessPoints;
}

QString SibConfig::getAccessPointParameters(QString configPointId)
{
    settings.beginGroup(CUTESIB_CONFIG_AP_PREFIX + configPointId);
    QString parameters = settings.value("parameters").toString();
    settings.endGroup();

    return parameters;
}

QString SibConfig::getAccessPointPath(QString accessPointId)
{
    settings.beginGroup(CUTESIB_CONFIG_AP_PREFIX + accessPointId);
    QString path = settings.value("fullpath").toString();

    if (path.isEmpty()) {
        path = settings.value("path").toString() + settings.value("name").toString() + SIB_AP_EXT;
    }

    settings.endGroup();

    if (path.isEmpty()) {
        path = CUTESIB_CONFIG_AP_DEFAULT_FOLDER;
    }

    return path;
}




