/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   09.03.2015
* @brief  !!!INSERT BRIEF!!!
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* !!!INSERT DESCRIPTION!!!.
**/

#include "ssaphandler.h"
#include "ssapparser.h"
#include "ssapmessage.h"
#include "ssapconstants.h"

#include <QDebug>


cutesib::SsapHandler::SsapHandler(SibConfig *config, QObject *parent) : QObject(parent)
{
    config_ = config;

    sib_data = sib_initialize("X", NULL);
}

cutesib::SsapHandler::~SsapHandler()
{
    librdf_free_model(sib_data->RDF_model);
    librdf_free_storage(sib_data->RDF_storage);
}




void cutesib::SsapHandler::addAccessPoint(access_point::AccessPointInterface *accessPoint)
{
    //TODO: nullptr or 0
    if (accessPoint == 0) {
        qWarning() << "incorrect access point: null pointer.";
        return;
    }

    // connect(accessPoint, &cutesib::access_point::AccessPointInterface::requestReceived, requestReceived);
    //  QObject::connect(accessPoint,
    //            &cutesib::access_point::AccessPointInterface::clientConnected,
    //            this,
    //           &cutesib::SsapHandler::clientConnected);

    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(clientConnected(cutesib::access_point::Client*)),
            this,
            SLOT(clientConnected(cutesib::access_point::Client*)));

    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(clientDisconnected(cutesib::access_point::Client*)),
            this,
            SLOT(clientDisconnected(cutesib::access_point::Client*)));

    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(requestReceived(cutesib::access_point::Client*, QString)),
            this,
            SLOT(requestReceived(cutesib::access_point::Client*, QString)));

    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(requestReceived(cutesib::access_point::Request*)),
            this,
            SLOT(requestReceived(cutesib::access_point::Request*)));


    connect(getIndcationMessnger(), &IndicatorMessanger::indicationReady,
            this, &SsapHandler::indicationHandle);

    accessPoints_.append(accessPoint);

}



void cutesib::SsapHandler::clientConnected(cutesib::access_point::Client *client)
{
    qDebug() << "Client connected.";
}

void cutesib::SsapHandler::requestReceived(cutesib::access_point::Client *client, QString data)
{
    //    NodeMsgContent_t *msg = parseSSAPmsg_new();
    //    client_receive_message(data.toUtf8().data(), msg);
    //    listener_handle__handle_request(msg, sib_data);



}

void cutesib::SsapHandler::requestReceived(cutesib::access_point::Request *request)
{
    SsapMessage message;
    message.load(request->getData());

    qDebug() << "SSAP message\n"
             << "\tTransaction type:" << message.transactionType() << message.transactionTypeText() << "\n"
             << "\tMessage type:" << message.messageType() << "\n"
             << "\tNeed confirmation" << message.isNeedConfirmation() << "\n"
             << "\tTransaction id:" << message.transactionId() << "\n"
             << "\tNode id:" << message.nodeId() << "\n"
             << "\tSpace id:" << message.spaceId() << "\n"
             << "\tInsert graph" << message.insertTriples().count();



    handleMessage(request, message);

    delete request;
}

void cutesib::SsapHandler::indicationHandle(subs_ind_t *ind)
{
    qDebug() << "Subscription indication for" << ind->node_id;

    foreach (Client *client, clientIds_.keys()) {
        if (clientIds_.value(client) == ind->subs_id) {
            qDebug() << "Send to client" << "All clients: " << clientIds_.count();
            client->getAccessPoint()->send(client, ind->ssap_message);
        }
    }
}

int cutesib::SsapHandler::start()
{
    using namespace cutesib::access_point;

    int failed_started = 0;

    foreach (AccessPointInterface *ap, accessPoints_) {
        if (ap->start() == -1) {
            qWarning() << "Starting access point failed:" << ap->configId();
            ++failed_started;
        } else {
            qDebug() << "Access point" << ap->configId() << "has started.";
        }
    }

    return failed_started;
}

void cutesib::SsapHandler::stop()
{
    using namespace cutesib::access_point;

    foreach (AccessPointInterface *ap, accessPoints_) {
        ap->stop();

        qDebug() << "Access point" << ap->configId() << "has stopped.";
    }
}

void cutesib::SsapHandler::clientDisconnected(cutesib::access_point::Client *client)
{
    QString subscriptionId = clientIds_.take(client);

    if (!subscriptionId.isEmpty()) {

        QString unsubscribeRequest = "<SSAP_message><transaction_type>UNSUBSCRIBE</transaction_type>"
                                     "<message_type>REQUEST</message_type>"
                                     "<node_id>" + client->nodeId() + "</node_id>"
                                     "<space_id>X</space_id><transaction_id>1</transaction_id>"
                                     "<parameter name = \"subscription_id\">" + subscriptionId + "</parameter>"
                                     "</SSAP_message>";

        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (unsubscribeRequest.toUtf8().data()), msg);

        char *response = listener_handle_unsubscribe(msg, sib_data);


        if (response != NULL) {
            qDebug() << "Subscription cancelled:" << response;
            free(response);
        } else {
            qWarning() << "Subscription cancelled ERROR:" << response;
        }

        parseSSAPmsg_free(&msg);

    }

    delete client;
}

void cutesib::SsapHandler::handleMessage(cutesib::access_point::Request *request, SsapMessage &message)
{
    if (message.messageType() != SsapConstants::MessageTypeRequest) {
        qDebug() << "Incorrect message: message type is not a request";
        //TODO: send error to client.
        return;
    }



    if (message.transactionType() == SsapTransactionType::SSAP_JOIN) {

        QString response = "<SSAP_message><message_type>CONFIRM</message_type><transaction_type>JOIN</transaction_type>"
                           "<transaction_id>" + message.transactionId() + "</transaction_id><space_id>" + message.spaceId() + "</space_id>"
                           "<node_id>" + message.nodeId() + "</node_id><parameter name=\"status\">m3:Success</parameter></SSAP_message>";


        request->getAccessPoint()->send(request->getClient(), response.toUtf8());


    }else if (message.transactionType() == SsapTransactionType::SSAP_LEAVE) {
        QString response = "<SSAP_message><message_type>CONFIRM</message_type><transaction_type>LEAVE</transaction_type>"
                           "<transaction_id>" + message.transactionId() + "</transaction_id><space_id>" + message.spaceId() + "</space_id>"
                           "<node_id>" + message.nodeId() + "</node_id><parameter name=\"status\">m3:Success</parameter></SSAP_message>";


        request->getAccessPoint()->send(request->getClient(), response.toUtf8());
    } else if (message.transactionType() == SsapTransactionType::SSAP_INSERT) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);
        //listener_handle_request(msg, sib_data);
        char *response = listener_handle_insert(msg, sib_data);

        if (response) {
            request->getAccessPoint()->send(request->getClient(), response);
            free(response);
        }

        parseSSAPmsg_free(&msg);
    } else if (message.transactionType() == SsapTransactionType::SSAP_QUERY) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);

        char *response = listener_handle_query(msg, sib_data);

        if (response) {
            request->getAccessPoint()->send(request->getClient(), response);
            free(response);
        }

        parseSSAPmsg_free(&msg);
    } else if (message.transactionType() == SsapTransactionType::SSAP_UPDATE) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);

        char *response = listener_handle_update(msg, sib_data);

        if (response) {
            request->getAccessPoint()->send(request->getClient(), response);
            free(response);
        }

        parseSSAPmsg_free(&msg);
    } else if (message.transactionType() == SsapTransactionType::SSAP_REMOVE) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);

        char *response = listener_handle_remove(msg, sib_data);

        if (response) {
            request->getAccessPoint()->send(request->getClient(), response);
            free(response);
        }

        parseSSAPmsg_free(&msg);
    } else if (message.transactionType() == SsapTransactionType::SSAP_SUBSCRIBE) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);

        char *response = listener_handle_subscribe(msg, sib_data);

        if (response) {
            request->getAccessPoint()->send(request->getClient(), response);
            //request->getClient()->setNodeId(message.nodeId());

            QRegExp regexpSubsId ("<parameter name=\"subscription_id\">(.*)");
            //regexpSubsId.setPatternSyntax(QRegExp::Wildcard);
            int pos = regexpSubsId.indexIn(response);
            if (pos > 0) {
                QStringList list = regexpSubsId.capturedTexts();
                int lastIndex = list.at(1).indexOf("</");
                QString subscriptionId = list.at(1).mid(0, lastIndex);

                ((Client*)request->getClient())->setNodeId(message.nodeId());
                clientIds_.insert(const_cast<Client *>(request->getClient()), subscriptionId);
                request->getAccessPoint()->startChecking(const_cast<Client*>(request->getClient()));

                qDebug() << "Subscribed clients" << clientIds_.values();
            }

            free(response);
        }

        parseSSAPmsg_free(&msg);
    } else if (message.transactionType() == SsapTransactionType::SSAP_UNSUBSCRIBE) {
        NodeMsgContent_t *msg = parseSSAPmsg_new();
        client_receive_message( const_cast<char*> (request->getData().data()), msg);

        char *response = listener_handle_unsubscribe(msg, sib_data);

        QString subscriptionId;
        QRegExp regexpSubsId ("<parameter name=\"subscription_id\">(.*)");
        int pos = regexpSubsId.indexIn(response);
        if (pos > 0) {
            QStringList list = regexpSubsId.capturedTexts();
            int lastIndex = list.at(1).indexOf("</");
            subscriptionId = list.at(1).mid(0, lastIndex);
            qDebug() << "Unsubscription for Id:" << subscriptionId;
        }


        if (response) {
            qDebug() << "Try found subscriber:" << subscriptionId << ". Subscribers:" << clientIds_.count();
            foreach (Client *client, clientIds_.keys()) {
                if (clientIds_.value(client) == subscriptionId) {
                    request->getAccessPoint()->stopChecking(client);
                    request->getAccessPoint()->send(client, response);
                    clientIds_.remove(const_cast<Client *>(client));
                    qDebug() << "Subscriber removed:" << subscriptionId << ". Subscribers:" << clientIds_.count();
                    break;
                }
            }

            free(response);
        }

        parseSSAPmsg_free(&msg);
    }
}
