/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   09.03.2015
* @brief  Handler to implement SSAP protocol.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Implementation of SSAP/XML protocol for SIB.
**/

#ifndef SSAPHANDLER_H
#define SSAPHANDLER_H

#include <QObject>

#include "client.h"
#include "accesspointinterface.h"
#include <whiteboard/sibmsg.h>

#include "ssapmessage.h"
#include "sibconfig.h"
#include "client.h"

#include "sib_data.h"

#include "sib_operations.h"
#include "command.h"

namespace cutesib {

/**
 * @brief The SsapHandler class implements SSAP/XML protocol.
 */
class SsapHandler : public QObject
{
    Q_OBJECT
public:
    explicit SsapHandler(SibConfig *config, QObject *parent = 0);
    ~SsapHandler();

    void addAccessPoint(access_point::AccessPointInterface *accessPoint);

    int start();
    void stop();


signals:
    void commandReady(Command *command);

public slots:
    // void errorOccured(QString data);
    void clientDisconnected(cutesib::access_point::Client* client);
    void clientConnected(cutesib::access_point::Client *client);
    void requestReceived(cutesib::access_point::Client *client, QString data);
    void requestReceived(cutesib::access_point::Request *request);

    void indicationHandle(subs_ind_t *ind);


private:
    void startHandleIndication();
    void stopHandleIndication();
    void handleIndication();



    void handleMessage(cutesib::access_point::Request *request, SsapMessage &message);
    //void hadleInsert(cutesib::o)
    QList<cutesib::access_point::AccessPointInterface*> accessPoints_;

    SibConfig *config_;

    sib_data_structure* sib_data;

    QList<cutesib::access_point::Client*> clients_;
    QMap<cutesib::access_point::Client*, QString> clientIds_;
};

} // namespace cutesib

#endif // SSAPHANDLER_H
