#include "ssapmessage.h"

#include <QXmlStreamReader>
#include <QDebug>

#include "ssapconstants.h"

#define CUTESIB_SSAP_TAG_TRANSACTION_TYPE "transaction_type"
#define CUTESIB_SSAP_TAG_MESSAGE_TYPE "message_type"
#define CUTESIB_SSAP_TAG_TRANSACTION_ID "transaction_id"
#define CUTESIB_SSAP_TAG_NODE_ID "node_id"
#define CUTESIB_SSAP_TAG_SPACE_ID "space_id"
#define CUTESIB_SSAP_TAG_PARAMETER "parameter"

#define CUTESIB_SSAP_TAG_TRIPLE_LIST "triple_list"
#define CUTESIB_SSAP_TAG_TRIPLE "triple"

#define CUTESIB_SSAP_TAG_PARAMETER "parameter"


#define CUTESIB_SSAP_TAG_ATTRIBUTE_CONFIRM "confirm"


cutesib::SsapMessage::SsapMessage(QObject *parent) : QObject(parent)
{
    xmlDoc = 0;
}

cutesib::SsapMessage::~SsapMessage()
{
    delete xmlDoc;
}

bool cutesib::SsapMessage::load(QByteArray message)
{
    delete xmlDoc;

    xmlDoc = new QXmlStreamReader(message);

    while (!xmlDoc->atEnd() && !xmlDoc->hasError()) {

        xmlDoc->readNext();

        if (xmlDoc->isEndElement()) {
            continue;
        }

        if (xmlDoc->name() == CUTESIB_SSAP_TAG_TRANSACTION_TYPE) {
            transactionType_ = convertTrsansactionType(xmlDoc->readElementText());
        } else if (xmlDoc->name() == CUTESIB_SSAP_TAG_MESSAGE_TYPE) {
            messageType_ = xmlDoc->readElementText();
        } else if (xmlDoc->name() == CUTESIB_SSAP_TAG_TRANSACTION_ID) {
            transactionId_ = xmlDoc->readElementText();
        } else if (xmlDoc->name() == CUTESIB_SSAP_TAG_NODE_ID) {
            nodeId_ = xmlDoc->readElementText();
        } else if (xmlDoc->name() == CUTESIB_SSAP_TAG_SPACE_ID) {
            spaceId_ = xmlDoc->readElementText();
        } else if (xmlDoc->name() == SsapConstants::TagParameter) {
            handleParameter(*xmlDoc);
        }
    }

    return true;
}





cutesib::SsapTransactionType cutesib::SsapMessage::transactionType()
{
    return transactionType_;
}

QString cutesib::SsapMessage::transactionTypeText()
{
    return transactionTypeText_;
}

QString cutesib::SsapMessage::messageType()
{
    return messageType_;
}
QString cutesib::SsapMessage::nodeId() const
{
    return nodeId_;
}
QString cutesib::SsapMessage::spaceId() const
{
    return spaceId_;
}
QString cutesib::SsapMessage::transactionId() const
{
    return transactionId_;
}

QString cutesib::SsapMessage::createResponse() const
{

}






cutesib::SsapTransactionType cutesib::SsapMessage::convertTrsansactionType(QString transactionType)
{
    transactionTypeText_ = transactionType;

    if (transactionType == SsapConstants::TransactionTypeJoin) {
        return SsapTransactionType::SSAP_JOIN;
    }

    if (transactionType == SsapConstants::TransactionTypeLeave) {
        return SsapTransactionType::SSAP_LEAVE;
    }

    if (transactionType == SsapConstants::TransactionTypeInsert) {
        return SsapTransactionType::SSAP_INSERT;
    }

    if (transactionType == SsapConstants::TransactionTypeQuery) {
        return SsapTransactionType::SSAP_QUERY;
    }

    if (transactionType == SsapConstants::TransactionTypeUpdate) {
        return SsapTransactionType::SSAP_UPDATE;
    }

    if (transactionType == SsapConstants::TransactionTypeRemove) {
        return SsapTransactionType::SSAP_REMOVE;
    }

    if (transactionType == SsapConstants::TransactionTypeSubscribe) {
        return SsapTransactionType::SSAP_SUBSCRIBE;
    }

    if (transactionType == SsapConstants::TransactionTypeUnsubscribe) {
        return SsapTransactionType::SSAP_UNSUBSCRIBE;
    }

    // Clean text if the transaction is unknown.
    transactionTypeText_.clear();

    return SsapTransactionType::SSAP_UNKNOWN;
}

void cutesib::SsapMessage::handleParameter(QXmlStreamReader &doc)
{

    QXmlStreamAttributes attributes = doc.attributes();

    qDebug() << "Attributes" << attributes.count() << attributes.value(SsapConstants::AttributeName);

    QStringRef name = attributes.value(SsapConstants::AttributeName);

    if (name == SsapConstants::AttributeValueConfirm) {
        isNeedConfirmation_ = true;
    } else if (name == SsapConstants::AttributeValueInsertGraph
               && attributes.value(SsapConstants::AttributeEncoding) == SsapConstants::AttributeValueRdfM3) {

        // Handle SSAP triples list (encoding="RDF-M3").
        doc.readNext();

        if (doc.name() == SsapConstants::TagTripleList) {
            insertTriples_ = handleTripleList(doc);
        } else {
            doc.readNextStartElement();
        }

    } else {
        doc.readNextStartElement();
    }


//    foreach (QXmlStreamAttribute attribute, doc.attributes()) {
//        if (attribute.name() == CUTESIB_SSAP_TAG_ATTRIBUTE_CONFIRM) {
//           isNeedConfirmation_ = true;
//        } else if (attribute.name() == CUTESIB_SSAP_TAG_ATTRIBUTE_INSERT_GRAPH) {
//            doc.readNext();

//            if (doc.name() == CUTESIB_SSAP_TAG_TRIPLE_LIST) {
//                qDebug() << doc.name();
//                if (doc.name() == CUTESIB_SSAP_TAG_TRIPLE) {
//                    //qDebug() <<
//                }

//            } else {
//                doc.readNextStartElement();
//            }

//        }
//    }


}

QList<cutesib::Triple*> cutesib::SsapMessage::handleTripleList(QXmlStreamReader &doc)
{
    QList<Triple*> triples;

    doc.readNext();

    while (!doc.atEnd() && !doc.hasError()) {

        if (doc.name() == SsapConstants::TagTriple
                && doc.isStartElement()) {
            Triple *triple = handleTriple(doc);

            if (triple) {
                triples.append(triple);
            }
        } else if (doc.name() == SsapConstants::TagTriple
                && doc.isEndElement()) {
            doc.readNext();
        } else {
            break;
        }
    }
    return triples;
}

cutesib::Triple* cutesib::SsapMessage::handleTriple(QXmlStreamReader &doc)
{
    QString subject = QString::null;
    QString object = QString::null;
    QString predicate = QString::null;
    RdfTripleType objectType = RdfTripleType::URI;

    doc.readNext();

    while (true) {
        qDebug() << doc.name() << doc.tokenString() << doc.attributes().count() << getAttribute(doc, "type") << getRdfType(doc);

        if (doc.name() == "subject") {
            subject = doc.readElementText();
        } else if (doc.name() == "object") {
            objectType = getRdfType(doc);
            object = doc.readElementText();
        } else if (doc.name() == "predicate") {
            predicate = doc.readElementText();
        } else {
            break;
        }

        doc.readNext();
    }

    if (subject.isNull()
            || predicate.isNull()
            || object.isNull()
            || objectType == RdfTripleType::UNKNOWN) {
        return 0;
    }

    return new Triple(subject, predicate, object, objectType);
}

QString cutesib::SsapMessage::getAttribute(QXmlStreamReader &doc, QString name)
{
    foreach (QXmlStreamAttribute attribute, doc.attributes()) {
        if (attribute.name() == name){
            return attribute.value().toString();
        }
    }

    return QString::null;
}

cutesib::RdfTripleType cutesib::SsapMessage::getRdfType(QXmlStreamReader &doc)
{
    QStringRef type = doc.attributes().value(SsapConstants::AttributeType); //getAttribute(doc, "type");

    if (type == SsapConstants::AttributeValueRdfTypeUri) {
        return RdfTripleType::URI;
    } else if (type == SsapConstants::AttributeValueRdfTypeLiteral) {
        return RdfTripleType::LITERAL;
    }

    return RdfTripleType::UNKNOWN;
}


bool cutesib::SsapMessage::isNeedConfirmation() const
{
    return isNeedConfirmation_;
}


const QList<cutesib::Triple *>  cutesib::SsapMessage::insertTriples() const
{
    return insertTriples_;
}
