#ifndef SSAPMESSAGE_H
#define SSAPMESSAGE_H

#include <QObject>
#include <QString>
#include <QXmlStreamReader>
#include <QList>

#include "triple.h"



namespace cutesib {

typedef enum SsapTransactionType {
    SSAP_UNKNOWN = -1,
    SSAP_JOIN,
    SSAP_LEAVE,
    SSAP_INSERT,
    SSAP_UPDATE,
    SSAP_REMOVE,
    SSAP_QUERY,
    SSAP_SUBSCRIBE,
    SSAP_UNSUBSCRIBE
} SsapTransactionType;


class SsapMessage : public QObject
{
    Q_OBJECT
public:
    explicit SsapMessage(QObject *parent = 0);
    ~SsapMessage();
    bool load(QByteArray message);
    SsapTransactionType transactionType();
    QString transactionTypeText();
    QString messageType();

    QString nodeId() const;

    QString spaceId() const;

    QString transactionId() const;

    QString createResponse() const;


    bool isNeedConfirmation() const;

    const QList<Triple *> insertTriples() const;




signals:

public slots:

private: // Methods
    SsapTransactionType convertTrsansactionType(QString transactionType);
    void handleParameter(QXmlStreamReader &doc);
    QList<Triple*> handleTripleList(QXmlStreamReader &doc);
    Triple *handleTriple(QXmlStreamReader &doc);
    QString getAttribute(QXmlStreamReader &doc, QString name);
    RdfTripleType getRdfType(QXmlStreamReader &doc);





private: // Fields
    QXmlStreamReader* xmlDoc;
    SsapTransactionType transactionType_;
    QString  transactionTypeText_;
    QString messageType_;
    QString transactionId_;
    QString nodeId_;
    QString spaceId_;
    bool isNeedConfirmation_ = false;

    QList<Triple*> insertTriples_;


};


}
#endif // SSAPMESSAGE_H
