/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   02.03.2015
* @brief  Access point that implements TCP protocol.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Access point that implements TCP protocol.
**/

#include <iostream>
#include <QObject>

#include "discoveryaccesspoint.h"


const QByteArray cutesib::access_point::DiscoveryAccessPoint::emptyString_ = QByteArray("\0");
const QString cutesib::access_point::DiscoveryAccessPoint::parameterPort_ = QString("port:");
const QString cutesib::access_point::DiscoveryAccessPoint::parameterSibPort_ = "sib_port:";
const QString cutesib::access_point::DiscoveryAccessPoint::parameterSibIp_ = "sib_ip:";
const QString cutesib::access_point::DiscoveryAccessPoint::requestAccessInfoText_ = "GET_SIB_ACCESS_INFORMATION";
const QString cutesib::access_point::DiscoveryAccessPoint::responseIncorrextRequestText_ = "error:incorrect_request";


cutesib::access_point::DiscoveryAccessPoint::DiscoveryAccessPoint(QObject *parent, int port) : QObject(parent)
{
    port_ = port;
    //TODO: change ip
    sibIp_ = "127.0.0.1";
    sibPort_ = 10010;
    connect(&socket_, SIGNAL(readyRead()), SLOT(readyRead()));
}

void cutesib::access_point::DiscoveryAccessPoint::setParameters(QString configId, QString parameters)
{
    configId_ = configId;

    QStringList parametersList = parameters.split(parametersDelimeter_);

    foreach (QString parameter, parametersList) {
        if (parameter.startsWith(parameterPort_)){
            setPort(parameter);
        } else if (parameter.startsWith(parameterSibIp_)) {
            setSibIp(parameter);
        } else if (parameter.startsWith(parameterSibPort_)) {
            setSibPort(parameter);
        }
    }
}

QString cutesib::access_point::DiscoveryAccessPoint::configId()
{
    return configId_;
}

int cutesib::access_point::DiscoveryAccessPoint::start()
{
    if (!socket_.bind(QHostAddress::Any, port_)) {
        qDebug() << "Discovery access point could not start.\n";
        return -1;
    }

    qDebug() << "Discovery access point started on port" << port_;

    return 0;
}

//TODO: check calling method on stopped server.
void cutesib::access_point::DiscoveryAccessPoint::stop()
{
    qDebug() << SIB_DEBUG_PREFIX << SIB_DEBUG_FNAME << endl;


}

void cutesib::access_point::DiscoveryAccessPoint::send(const cutesib::access_point::Client *client, const QByteArray data)
{

}

bool cutesib::access_point::DiscoveryAccessPoint::isConnected(const cutesib::access_point::Client *client)
{

}

void cutesib::access_point::DiscoveryAccessPoint::startChecking(cutesib::access_point::Client *client)
{

}

void cutesib::access_point::DiscoveryAccessPoint::stopChecking(cutesib::access_point::Client *client)
{

}

void cutesib::access_point::DiscoveryAccessPoint::close(cutesib::access_point::Client *client)
{

}




void cutesib::access_point::DiscoveryAccessPoint::readyRead()
{
    while (socket_.hasPendingDatagrams()) {
        QByteArray datagram;
        datagram.resize(socket_.pendingDatagramSize());
        QHostAddress sender;
        quint16 senderPort;

        if (socket_.readDatagram(datagram.data(), datagram.size(),
                                &sender, &senderPort) > 0) {
            QString request(datagram.data());

            qDebug() << "Request: " << request;
            if (request == requestAccessInfoText_) {
                qDebug() << sibPort_;
                QString responseText = parameterSibIp_ + sibIp_ + parametersDelimeter_ + parameterSibPort_ + QString(sibPort_).toUtf8();
                qDebug() << sibPort_;
                qDebug() << "Response: " << responseText << QString(sibPort_);
                send(sender, senderPort, responseText.toUtf8());
            } else {
                send(sender, senderPort, "error:incorrect_request");
            }
        }
    }



//    qDebug()<<"Datagramm recived";
//    QByteArray datagram;
//    datagram.resize(socket_.pendingDatagramSize());

//    QHostAddress address;
//    quint16 port;

//    socket_.readDatagram(datagram.data(), datagram.size(), &address, &port);

//    QDataStream in(&datagram, QIODevice::ReadOnly);

//    qint64 size = -1;

//    if(in.device()->size() > sizeof(qint64)) {
//        in >> size;
//    }

//    if (in.device()->size() - sizeof(qint64) < size) return;

//      QString clientData;
//      in >> clientData;

//      qDebug() << clientData;
//      qDebug() << "Send to" << address.toString() << port;
//      send(address, port, "sdaadsads");


}




bool cutesib::access_point::DiscoveryAccessPoint::send(QHostAddress clientAddress, quint16 clientPort, const QString message)
{
       QByteArray data;
       QDataStream out(&data, QIODevice::WriteOnly);

       out << qint64(0);
       out << message;
       out.device()->seek(qint64(0));
       out << qint64(data.size() - sizeof(qint64));

       if (socket_.writeDatagram(data, clientAddress, clientPort) <= 0) {
           return false;
       }

       return true;
}




void cutesib::access_point::DiscoveryAccessPoint::setPort(QString parameter)
{
    int port = parameter.remove(0, parameterPort_.length()).toInt();

    if (port > 0) {
        port_ = port;
    }
}

void cutesib::access_point::DiscoveryAccessPoint::setSibPort(QString parameter)
{
    int port = parameter.remove(0, parameterSibPort_.length()).toInt();

    qDebug() << "SIB port = " << port << sibPort_;
    if (port > 0) {
        sibPort_ = port;
    }
}

void cutesib::access_point::DiscoveryAccessPoint::setSibIp(QString parameter)
{
    QString ip = parameter.remove(0, parameterSibIp_.length());

    if (!ip.isEmpty()) {
        sibIp_ = ip;
    }
}


