/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   05.03.2015
* @brief  Class for TCP client.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Class for TCP client.
**/

#include "tcpclient.h"
#include "accesspointinterface.h"

#include <QBuffer>

namespace cutesib_ap = cutesib::access_point;

const QByteArray cutesib_ap::TcpClient::MessageEnding = "</SSAP_message>";

cutesib_ap::TcpClient::TcpClient(AccessPointInterface *accessPoint, QTcpSocket *socket)
{
    socket_ = socket;
    accessPoint_ = accessPoint;
//    setParent((QObject *)accessPoint);
    isConnected = true;
}

cutesib_ap::TcpClient::~TcpClient()
{
    if (socket_->isOpen()){
        socket_->close();
    }

    delete socket_;
}

QTcpSocket *cutesib::access_point::TcpClient::getSocket() const
{
    return socket_;
}

bool cutesib::access_point::TcpClient::isReceivedFullMessage()
{
    qDebug() << "Message has end tag:" << receiveBuffer_.contains(MessageEnding);
    return receiveBuffer_.contains(MessageEnding);
}

void cutesib::access_point::TcpClient::addMessagePart(QByteArray messagePart)
{
    qDebug() << "Buffer:" << receiveBuffer_;
    receiveBuffer_.append(messagePart);
    qDebug() << "New buffer:" << receiveBuffer_;
}

QByteArray cutesib::access_point::TcpClient::getMessage()
{

    int index = receiveBuffer_.indexOf(MessageEnding);
    index += MessageEnding.length();

    QByteArray temp;
    temp.append(receiveBuffer_.data(), index);

    receiveBuffer_ = receiveBuffer_.remove(0, index);



    return temp;
}






cutesib::access_point::AccessPointInterface *cutesib::access_point::TcpClient::getAccessPoint()
{
    return accessPoint_;
}

void cutesib::access_point::TcpClient::Lock()
{
    socketLock_.lock();
}

void cutesib::access_point::TcpClient::Unlock()
{
    socketLock_.unlock();
}

void cutesib::access_point::TcpClient::setMessage(QByteArray data)
{
    if (messageBuffer.isOpen()) {
        messageBuffer.close();
    }

    messageBuffer.setData(data);

    messageBuffer.open(QIODevice::ReadOnly);
}

qint64 cutesib::access_point::TcpClient::messageSize()
{
    return messageBuffer.size();
}

QByteArray cutesib::access_point::TcpClient::getMessagePart(qint64 size)
{
    if (!messageBuffer.isOpen()) {
        messageBuffer.open(QIODevice::ReadOnly);
    }

    return messageBuffer.read(size);
}

