/**
* @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
* @date   05.03.2015
* @brief  Class for TCP client.
*
* This file is part of the QT-SIB project.
*
* Copyright (C) 2015 by PetrSU.
* Contact: http://oss.fruct.org/wiki/SmartSpace-platform
*
* @section LICENSE
*
* QT-SIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* QT-SIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with QT-SIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Class for TCP client.
**/

#ifndef TCPCLIENT_H
#define TCPCLIENT_H

#include "tcpaccesspoint_global.h"

#include <QObject>
#include <QBuffer>
#include <QMutex>
#include <QTcpSocket>

#include "client.h"
#include "accesspointinterface.h"

namespace cutesib {
namespace access_point {

/**
 * @brief The TcpClient class represents a client for TCP access point.
 * @see Client
 */
class TcpClient : public Client
{
    Q_OBJECT

public:
    TcpClient(AccessPointInterface *accessPoint, QTcpSocket* socket);
    ~TcpClient();

    static const QByteArray MessageEnding;

    QTcpSocket* getSocket() const;
    bool isConnected;

    bool isReceivedFullMessage();

    void addMessagePart(QByteArray messagePart);

    QByteArray getMessage();

    AccessPointInterface *getAccessPoint();

    void Lock();
    void Unlock();

    void setMessage(QByteArray data);
    qint64 messageSize();
    QByteArray getMessagePart(qint64 size);

private:
    QTcpSocket* socket_;
    QByteArray receiveBuffer_;
    AccessPointInterface *accessPoint_;
    QMutex socketLock_;
    QBuffer messageBuffer;
};

} // namespace access_point
} // namespace cutesib

#endif // TCPCLIENT_H
